import os

import histoprep
import pytest
from PIL import Image

from paths import DATA_PATH

SLIDE_1 = os.path.join(DATA_PATH, "slide_1.svs")
SLIDE_2 = os.path.join(DATA_PATH, "slide_2.mrxs")
SLIDE_3 = os.path.join(DATA_PATH, "slide_3.czi")
SLIDE_4 = os.path.join(DATA_PATH, "tile.jpeg")
OUTPUT_DIR = os.path.join(DATA_PATH, "output")
SCRATCH_DIR = os.path.join(DATA_PATH, "scratch")


def test_bad_paths():
    with pytest.raises(FileNotFoundError):
        histoprep.SlideReader("i/dont/exist")
    with pytest.raises(IOError):
        histoprep.SlideReader(DATA_PATH)


def test_bad_region():
    reader = histoprep.SlideReader(os.path.join(DATA_PATH, "slide_1.svs"))
    for xywh in [
        (-1, 0, 10, 10),
        (0, -1, 10, 10),
        (0, 0, 0, 10),
        (0, 0, 10, -1),
    ]:
        with pytest.raises(ValueError):
            reader.read_region(xywh)


def test_out_of_bounds():
    reader = histoprep.SlideReader(os.path.join(DATA_PATH, "slide_1.svs"))
    with pytest.raises(ValueError):
        reader.read_region((reader.dimensions[1] + 10, 0, 10, 10), fill=None)


def test_get_thumbnail():
    reader = histoprep.SlideReader(os.path.join(DATA_PATH, "slide_1.svs"))
    for kwargs in [
        dict(preferred_dimension=None, level=None),
        dict(preferred_dimension=None, level=1000),
    ]:
        with pytest.raises(ValueError):
            reader.get_thumbnail(**kwargs)
    # Successes.
    assert reader.get_thumbnail(level=2).size == reader.level_dimensions[2][::-1]
    assert reader.get_thumbnail(2187).size == reader.level_dimensions[2][::-1]
    assert reader.get_thumbnail(1).size == reader.level_dimensions[2][::-1]


# def test_scratch():
#     reader = histoprep.SlideReader(SLIDE_1, scratch_dir=SCRATCH_DIR)
#     assert os.path.exists(os.path.join(SCRATCH_DIR, "slide_1.svs"))
#     shutil.rmtree(os.path.join(DATA_PATH, "scratch"))


def test_properties():
    reader = histoprep.SlideReader(SLIDE_1)
    assert str(reader.backend) == "OPENSLIDE"
    assert reader.channel_order == "XYWH"
    assert reader.dimension_order == "HW"
    assert isinstance(reader.dimensions, tuple) and isinstance(
        reader.dimensions[0], int
    )
    assert isinstance(reader.level_dimensions, dict)
    assert reader.level_dimensions[0] == reader.dimensions
    assert isinstance(reader.level_downsamples, dict)
    assert reader.level_downsamples[0] == (1, 1)
    assert reader.path == SLIDE_1
    assert reader.slide_name == "slide_1"
    assert isinstance(reader.thumbnail, Image.Image)
    assert [int(x) for x in reader.thumbnail_downsample] == [8, 8]
    assert reader.tile_metadata is None
    assert isinstance(reader.tissue_mask, Image.Image)
    assert reader.tissue_mask.size == reader.thumbnail.size
    # Initialized properties with no data.
    assert reader.annotated_thumbnail_spots is None
    assert reader.annotated_thumbnail_tiles is None
    assert reader.spot_mask is None
    assert reader.spot_metadata is None


def test_tissue_detection():
    reader = histoprep.SlideReader(SLIDE_1)
    thresh, mask = reader.detect_tissue(threshold=0, multiplier=1.0)
    assert mask.sum() == 0
    assert thresh == reader.tissue_threshold
    thresh, mask = reader.detect_tissue(threshold=255, multiplier=1.0)
    assert mask.sum() == mask.size
    assert thresh == reader.tissue_threshold


def test_max_dimension():
    with pytest.raises(ValueError):
        histoprep.SlideReader(SLIDE_1, max_dimension=200)
    reader = histoprep.SlideReader(SLIDE_1)
    assert reader.MAX_DIMENSION == 2**14
    with pytest.raises(TypeError):
        reader.MAX_DIMENSION = "penis"
    with pytest.raises(ValueError):
        reader.MAX_DIMENSION = -1
    reader.MAX_DIMENSION = 4000



import time
from typing import Iterable

__all__ = ["progress_bar"]


def progress_bar(
    iterable: Iterable,
    total: int = None,
    desc: str = None,
    log_interval: int = 1,
    log_values: bool = False,
    suppress: bool = False,
):
    """Simple print-based progress bar.

    Args:
        iterable: Iterable to wrap.
        total: Total steps. Defaults to None.
        desc: Description for progress bar. Defaults to None.
        log_interval: Log every n steps. Defaults to 1.
        log_values: Returns (dict, next(iterable)), where values added to the dict are
            logged to the progress bar. Defaults to True.
        suppress: Suppress all output. Defaults to False.

    Yields:
        Iterable output
    """
    if suppress:
        for output in iterable:
            yield output
        return
    try:
        total = len(iterable) if total is None else total
    except TypeError:
        total = None
    tic = time.perf_counter()
    tic_out = time.perf_counter()
    logs = {}
    try:
        bar = __get_bar(0, tic, total, desc=desc, logs=logs)
        last_length = len(bar)
        print(bar, end="\r")
        for step, output in enumerate(iterable):
            if step == 0 or (
                (step + 1) % log_interval == 0 and time.perf_counter() - tic_out > 0.05
            ):
                bar = __get_bar(step + 1, tic, total, desc=desc, logs=logs)
                print(" " * last_length, end="\r")
                print(bar, end="\r")
                last_length = len(bar)
                tic_out = time.perf_counter()
            if log_values:
                yield logs, output
            else:
                yield output
        print(" " * last_length, end="\r")
        print(__get_bar(step + 1, tic, total, desc=desc, final=True, logs=logs))
    except KeyboardInterrupt:
        print("\n Keyboard interruption detected.")


def __get_bar(
    step: int,
    tic: float,
    total: int = None,
    bar_length: int = 10,
    final: bool = False,
    desc: str = None,
    logs: dict = {},
):
    elapsed = __format_seconds(time.perf_counter() - tic)
    log_str = ""
    if len(logs) > 0:
        log_str = ", " + ", ".join(["{}={}".format(k, v) for k, v in logs.items()])
    if total is None:
        prefix = "" if desc is None else "{}: ".format(desc)
        bar = "{}{}it [{}{}]".format(prefix, step, elapsed, log_str)
    else:
        n = min(bar_length, int(bar_length * step / total))
        if final:
            bar = "#" * bar_length
        else:
            bar = "#" * n + " " * (bar_length - n)
        if step == 0:
            etc = "???"
        else:
            s_per_iter = (time.perf_counter() - tic) / step
            etc = __format_seconds((total - step) * s_per_iter)
        prefix = "" if desc is None else desc + " "
        bar = "{}|{}| {}/{} [{}<{}{}]".format(
            prefix, bar, step, total, elapsed, etc, log_str
        )
    return bar


def __format_seconds(n: int):
    """Format seconds into pretty string format."""
    days = int(n // (24 * 3600))
    n = n % (24 * 3600)
    hours = int(n // 3600)
    n %= 3600
    minutes = int(n // 60)
    n %= 60
    seconds = int(n)
    if days > 0:
        strtime = f"{days}d {hours}h:{minutes:02}m:{seconds:02}s"
    elif hours > 0:
        strtime = f"{hours}:{minutes:02}:{seconds:02}"
    else:
        strtime = f"{minutes:02}:{seconds:02}"
    return strtime


def verbose_fn(msg: str, desc: str, verbose: bool = True, color: bool = True):
    """Verbose output with description."""
    if not verbose:
        return
    if color:
        print("\x1b[1m\x1b[34m[{}]\x1b[0m {}".format(desc, msg))
    else:
        print("[{}] {}".format(desc, msg))



import os

import numpy
import pandas
import pytest
from histoprep import OutlierDetector, OutlierVisualizer

from paths import DATA_PATH

metadata = pandas.read_csv(os.path.join(DATA_PATH, "metadata.csv"))
visual = OutlierVisualizer(metadata)
detect = OutlierDetector(metadata)


def test_visualiser():
    visual.plot_rgb_std(4)
    visual.plot_rgb_std(4, log_scale=True)
    visual.plot_hsv_std(4)
    visual.plot_hsv_std(4, log_scale=True)
    visual.plot_rgb_quantiles(4)
    visual.plot_rgb_quantiles(4, log_scale=True)
    visual.plot_hsv_quantiles(4)
    visual.plot_hsv_quantiles(4, log_scale=True)


def test_detector():
    assert len(detect.cluster_counts) == detect.num_clusters
    assert len(detect.cluster_distances) == detect.num_clusters
    assert isinstance(detect.metrics, numpy.ndarray)
    assert all(numpy.isclose(detect.metrics.mean(0), 0))
    assert len(detect.clusters) == detect.metrics.shape[0]
    detect.plot_clusters(10)
    detect.plot_cluster(0)
    with pytest.warns(UserWarning):
        detect.plot_cluster(100000)


def test_umap_repr():
    with pytest.warns():
        detect.umap_representation(2, max_samples=100)


def test_pca_repr():
    coords, indices = detect.pca_representation(2, max_samples=100)
    assert coords.shape == (100, 2)
    assert indices.size == 100


def test_plot_repr():
    coords, indices = detect.pca_representation(2, max_samples=100)
    detect.plot_representation(coords, indices)



from histoprep.helpers._time import format_seconds, get_ETC


def test_format_seconds():
    for dst, src in {
        "59s": 59,
        "1m:0s": 60,
        "1h:0m:1s": 60 * 60 + 1,
        "1d 0h:0m:1s": 60 * 60 * 24 + 1,
        "1y 0d 0h:0m:1s": 60 * 60 * 24 * 365.25 + 1,
    }.items():
        assert dst == format_seconds(src)


def test_ETC():
    times = [1, 2]
    assert "ETC: 3s" == get_ETC(times, 2)
    assert "ETC: 4.5s" == get_ETC(times, 3)
    assert "" == get_ETC([], 2)



import math
import random
import warnings
from typing import Dict, Tuple

import matplotlib.pyplot as plt
import numpy
import pandas
from PIL import Image
from sklearn.cluster import MiniBatchKMeans
from sklearn.decomposition import PCA

from ..helpers._io import read_and_resize
from ..helpers._visualize import random_tile_collage
from ._visualize import FIGURE_HEIGHT, FIGURE_WIDTH

__all__ = ["OutlierDetector"]


class OutlierDetector:
    def __init__(
        self,
        metadata: pandas.DataFrame,
        num_clusters: int = 20,
        batch_size: int = 2**14,
        **kwargs,
    ):
        """Automatic outlier detection from color channel quantile metrics.

        Kmeans++ is used to cluster each tile into the desired number of
        clusters. Clusters are ordered by the distance to the origo in
        decreasing order, and thus the most likely outlier group is first and
        last groups contain normal images.

        Args:
            metadata: Metadata containing columns: `[channel_name]_q={quantile}`
            num_clusters: Number of clusters. Defaults to 20.
            batch_size: Batch size for kmeans++. Defaults to 2**14.

        Example:
            ```python
            import histoprep

            detector = histoprep.OutlierDetector(metadata, num_clusters=20)
            detect.plot_clusters(min_distance=10)
            # From the plots we might see that clusters 0-3 contain outliers.
            metadata["outlier"] = False
            metadata.loc[detector.clusters < 4, "outlier"] = True
            ```
        """
        self.__num_clusters = num_clusters
        # Pull paths from metadata.
        if "path" not in metadata.columns:
            raise ValueError("Metadata should not contain a 'path' column.")
        self.__all_paths = metadata["path"].to_numpy()
        # Then quantile metrics.
        X = metadata[[x for x in metadata.columns if "_q=" in x]].to_numpy()
        X = (X - X.mean(axis=0)) / X.std(axis=0)
        self.__metrics = X
        # Cluster.
        self.__clusters = self.__cluster_data(
            X,
            n_clusters=num_clusters,
            batch_size=batch_size,
            **kwargs,
        )

    @property
    def metrics(self) -> numpy.ndarray:
        """Quantile metrics normalised with: ``(X - X.mean(0)) / X.std(0)``"""
        return self.__metrics

    @property
    def num_clusters(self) -> int:
        return self.__num_clusters

    @property
    def clusters(self) -> numpy.ndarray:
        """Cluster assignments."""
        return self.__clusters["clusters"]

    @property
    def cluster_counts(self) -> numpy.ndarray:
        """Number of members in each cluster."""
        return self.__clusters["counts"]

    @property
    def cluster_distances(self) -> numpy.ndarray:
        """Cluster mean euclidean distance to the origo."""
        return self.__clusters["distances"]

    @staticmethod
    def __cluster_data(X: numpy.ndarray, **kwargs) -> Dict[str, numpy.ndarray]:
        """Helper function to cluster metrics."""
        # Define cluster instance.
        kmeans = MiniBatchKMeans(**kwargs)
        # Cluster, get counts and calculate distance from origo.
        clusters = kmeans.fit_predict(X)
        __, counts = numpy.unique(clusters, return_counts=True)
        distances = numpy.array([numpy.linalg.norm(x) for x in kmeans.cluster_centers_])
        # Rename starting from the most distant.
        new_clusters = numpy.zeros_like(clusters)
        for new_idx, old_idx in enumerate(distances.argsort()[::-1]):
            new_clusters[clusters == old_idx] = new_idx
        # Cache clusters.
        return {
            "clusters": new_clusters,
            "counts": counts[distances.argsort()[::-1]],
            "distances": distances[distances.argsort()[::-1]],
        }

    def plot_clusters(
        self,
        min_distance: int = None,
        num_examples: int = 256,
        ncols: int = 32,
        px: int = 64,
        figsize: Tuple[float, float] = (FIGURE_WIDTH * 1.5, FIGURE_HEIGHT),
    ):
        """Plot example images from clusters.

        Args:
            min_distance: Minimum distance to filter which groups to plot.
                Defaults to None.
            num_examples: Number of example images from each group. Defaults to 256.
            ncols: Number of columns for plotting example images. Defaults to 32.
            px: Size of each example image. Defaults to 64.
            figsize: Figure size for plotting. Defaults to
                (default_width * 1.5, default_height)
        """
        # Collect all paths.
        paths = []
        for idx, dist in enumerate(self.cluster_distances):
            if min_distance is not None and dist < min_distance:
                break
            # Load cluster paths and select num_examples.
            cluster_paths = self.__all_paths[self.clusters == idx]
            if len(cluster_paths) < num_examples:
                cluster_paths = cluster_paths.tolist()
                cluster_paths += [None] * (num_examples - len(cluster_paths))
            else:
                cluster_paths = numpy.random.choice(
                    cluster_paths, size=num_examples
                ).tolist()
            paths += cluster_paths
        # Start reading images and plot as they are loaded.
        row = []
        collage = []
        cluster_idx = 0
        for tile in read_and_resize(paths, px):
            row.append(tile)
            if len(row) == ncols:
                collage.append(numpy.hstack(row))
                row = []
            if len(collage) == math.ceil(num_examples / 32):
                # Create collage and plot.
                collage = Image.fromarray(numpy.vstack(collage))
                plt.figure(figsize=figsize)
                plt.imshow(collage)
                plt.title(
                    "Cluster {} (images={} - distance={:.3f})".format(
                        cluster_idx,
                        self.cluster_counts[cluster_idx],
                        self.cluster_distances[cluster_idx],
                    )
                )
                plt.axis("off")
                plt.tight_layout()
                plt.show()
                # Reset collage.
                collage = []
                cluster_idx += 1

    def plot_cluster(
        self,
        cluster: int = 0,
        num_examples: int = 256,
        ncols: int = 32,
        px: int = 64,
        figsize: Tuple[float, float] = (FIGURE_WIDTH * 1.5, FIGURE_HEIGHT),
    ):
        """Plot example images from a single cluster.

        Args:
            cluster: Cluster number. Defaults to 0.
            num_examples: Number of example images from each group. Defaults to 256.
            ncols: Number of columns for plotting example images. Defaults to 32.
            px: Size of each example image. Defaults to 64.
            figsize: Figure size for plotting. Defaults to
                (default_width * 1.5, default_height)
        """
        # Collect paths.
        cluster_paths = self.__all_paths[self.clusters == cluster]
        if len(cluster_paths) == 0:
            warnings.warn("No image paths found for cluster {}.".format(cluster))
            return
        # Plot.
        plt.figure(figsize=figsize)
        plt.imshow(
            random_tile_collage(
                paths=cluster_paths,
                nrows=num_examples // ncols + 1,
                ncols=ncols,
                px=px,
            )
        )
        plt.title(
            "Cluster {} (images={} - distance={:.3f})".format(
                cluster,
                self.cluster_counts[cluster],
                self.cluster_distances[cluster],
            )
        )
        plt.axis("off")
        plt.tight_layout()
        plt.show()

    def umap_representation(
        self,
        num_components: int = 2,
        num_neighbors: int = 15,
        metric: str = "euclidean",
        max_samples: int = 100_000,
        verbose: bool = True,
        **kwargs,
    ) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """Run Uniform Manifold Approximation and Projection (UMAP)
        dimensionality reduction.

        Args:
            num_components: Number of components, two offers easy visualisation.
                Defaults to 2.
            num_neighbors: Number of neighbours, should be between 2-100.
                Defaults to 15.
            max_samples: UMAP does not handle very large sample sizes well.
                Increase if you want to go and get coffee. Defaults to 100_000.
            **kwargs: Any keyword arguments for `umap.UMAP`.

        Returns:
            UMAP representation and indices of the selected samples
        """
        try:
            import umap
        except ImportError:
            warnings.warn(
                "Could not import `umap`. Please run `pip install umap-learn` "
                "to use this function."
            )
            return
        # Initialize.
        reducer = umap.UMAP(
            n_components=num_components,
            n_neighbors=num_neighbors,
            metric=metric,
            verbose=verbose,
            **kwargs,
        )
        # Fit and transform.
        if self.__metrics.shape[0] > max_samples:
            indices = numpy.array(
                random.choices(range(self.__metrics.shape[0]), k=max_samples)
            )
        else:
            indices = numpy.arange(self.__metrics.shape[0])
        representation = reducer.fit_transform(self.__metrics[indices, :])
        return representation, indices

    def pca_representation(
        self,
        num_components: int = 2,
        max_samples: int = 1_000_000,
        **kwargs,
    ) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """Run princical component analysis (PCA) to get a representation of
        metrics. Does not produce as nice results as UMAP, but it's faster and
        supported out-of-the-box.

        Args:
            num_components: Number of components, two offers easy visualisation.
                Defaults to 2.
            max_samples: PCA does handle very large sample sizes well, but this argument
                can be used to limit the number of samples. Defaults to 1_000_000.
            **kwargs: Any keyword arguments for `sklearn.decomposition.PCA`.

        Returns:
            PCA representation and indices of the selected samples
        """
        # Initialize.
        reducer = PCA(n_components=num_components, **kwargs)
        # Fit and transform.
        if self.__metrics.shape[0] > max_samples:
            indices = numpy.array(
                random.choices(range(self.__metrics.shape[0]), k=max_samples)
            )
        else:
            indices = numpy.arange(self.__metrics.shape[0])
        representation = reducer.fit_transform(self.__metrics[indices, :])
        return representation, indices

    def plot_representation(
        self,
        coordinates: numpy.ndarray,
        indices: numpy.ndarray,
        label_clusters: bool = False,
        size: float = 0.5,
        cmap: str = None,
        figsize: Tuple[float, float] = (
            FIGURE_WIDTH * 1.5,
            FIGURE_HEIGHT * 1.5,
        ),
    ):
        """Plot representation of metrics.

        Args:
            coordinates: Representation coordinates.
            indices: Sample indices returned by `{umap, pca}_representation`.
            label_clusers: Label each cluster in the plot instead of log(distance).
            size: Size of each dot. Defaults to 0.5.
            cmap: Colour map. Defaults to "plasma" for distance and "coolwarm_r" for
                clusters.
            figsize: Figure size. Defaults to  Defaults to
                (default_width * 1.5, default_height * 1.5).
        """
        if label_clusters:
            labels = self.clusters[indices]
            cmap = "coolwarm_r" if cmap is None else cmap
        else:
            labels = numpy.log(numpy.linalg.norm(self.metrics[indices], axis=1))
            cmap = "plasma" if cmap is None else cmap
        # Plot
        fig, ax = plt.subplots(figsize=figsize)
        scatter = plt.scatter(
            x=coordinates[:, 0],
            y=coordinates[:, 1],
            s=size,
            cmap=cmap,
            c=labels,
            label=labels if label_clusters else None,
        )
        if label_clusters:
            ax.legend(
                *scatter.legend_elements(num=self.num_clusters - 1),
                title="Clusters",
                ncol=2,
            )
        else:
            plt.colorbar(label="log(distance)", shrink=0.75)
        ax.set_title("Representation of all tile metrics.")
        plt.axis("off")
        plt.tight_layout()
        plt.show()

    def __repr__(self):
        repr_str = "{}(num_clusters={}):\n".format(
            self.__class__.__name__, self.num_clusters
        )
        for idx, (count, dist) in enumerate(
            zip(self.cluster_counts, self.cluster_distances)
        ):
            repr_str += "{:>4}:  dist={:<7.2f} images={}\n".format(idx, dist, count)
            if idx > 9:
                repr_str += "{:>10}".format("...")
                break
        return repr_str



import cv2
import numpy


def _thresholding(
    gray: numpy.ndarray,
    threshold: int = None,
    multiplier: float = 1.0,
    remove_white: bool = False,
) -> numpy.ndarray:
    if gray.ndim != 2:
        raise ValueError("Thresholding expects grayscale images.")
    if threshold is None:
        # Use Otsu.
        if remove_white:
            # Get threshold without white pixels.
            threshold, __ = cv2.threshold(
                gray[gray != 255],
                None,
                1,
                cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU,
            )
        else:
            # Normal thresholding.
            threshold, __ = cv2.threshold(
                gray, None, 1, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU
            )
        # Multiply threshold.
        threshold = max(min(255, int(threshold * multiplier)), 0)
    # Global thresholding.
    threshold, mask = cv2.threshold(gray, threshold, 1, cv2.THRESH_BINARY_INV)
    return int(threshold), mask


def _gaussian_blur(
    image: numpy.ndarray, sigma: float, truncate: float = 3.5
) -> numpy.ndarray:
    """Gaussian blur."""
    # Define kernel size based on sigma and truncate.
    if sigma > 0:
        ksize = int(truncate * sigma + 0.5)
        if ksize % 2 == 0:
            ksize += 1
        kernel_size = (ksize, ksize)
        # Gaussian blurrrr.
        return cv2.GaussianBlur(image, ksize=kernel_size, sigmaX=sigma, sigmaY=sigma)
    else:
        return image


def _to_grayscale(image: numpy.ndarray) -> numpy.ndarray:
    """Convert image to grayscale."""
    if image.ndim == 2:
        return image
    else:
        return cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)


def _to_laplacian(gray: numpy.ndarray) -> numpy.ndarray:
    """Apply laplacian filter to the image."""
    if gray.ndim != 2:
        gray = _to_grayscale(gray)
    return cv2.Laplacian(gray, cv2.CV_32F)


def _rgb2hsv(image: numpy.ndarray) -> numpy.ndarray:
    """Convert RGB to HSV."""
    return cv2.cvtColor(image, cv2.COLOR_RGB2HSV)



from typing import Callable, Dict, List, Tuple, Union

import numpy
from PIL import Image

from ._functional import _rgb2hsv, _to_grayscale, _to_laplacian
from ._helpers import _five_crop, _reduce_values, _to_array, resize_image
from ._tissue import detect_tissue

DEFAULT_QUANTILES = [0.05, 0.1, 0.5, 0.9, 0.95]

__all__ = [
    "PreprocessMetrics",
    "data_loss",
    "sharpness",
    "channel_quantiles",
    "channel_std",
]


class PreprocessMetrics:
    """Class to calculate basic preprocessing metrics for a histological image.

    Args:
        channel_resize: Image size for `F.channel_quantiles()`. Defaults to 128.
        quantiles: RGB and HSV quantiles. Defaults to
            [0.05, 0.1, 0.5, 0.9, 0.95].
        sharpness_reduce: Sharpness reduction method. Defaults to "max".
        custom_callback: Custom callback for to calculate new metrics. Defaults
            to None.

    Example:
        ```python
        import histoprep.functional as F
        from histoprep.helpers import read_image

        image = read_image("path/to/image.jpeg")
        threshold, tissue_mask = F.detect_tissue(image)
        metric_fn = PreprocessMetrics(threshold)
        results = metric_fn(image)
        ```
    """

    def __init__(
        self,
        channel_resize: int = 64,
        quantiles: List[int] = DEFAULT_QUANTILES,
        sharpness_reduce: Union[str, list] = "max",
        custom_callback: Callable = None,
    ):
        self.__custom_callback = custom_callback
        self.__quantiles = quantiles
        self.__channel_resize = channel_resize
        self.__sharpness_reduce = sharpness_reduce

    def __call__(
        self,
        image: Union[numpy.ndarray, Image.Image],
        tissue_threshold: int,
    ) -> Dict[str, float]:
        """Calculate preprocessing metrics for an image,

        Args:
            image: Input image.
            threshold: Tissue detection threshold for the tile images. Otsu's
                method cannot not be used here as some of the tiles may not
                contain background, and the method would still find an "optimal"
                threshold and classify tissue as background.

        Returns:
            Preprocessing metrics.
        """
        # Convert to array.
        image = _to_array(image)
        # Detect tissue.
        __, tissue_mask = detect_tissue(image, threshold=tissue_threshold)
        # Initialize metrics with background.
        preprocessing_metrics = {
            "background": (tissue_mask == 0).sum() / tissue_mask.size
        }
        # Channel mean and std.
        preprocessing_metrics.update(channel_mean(image))
        preprocessing_metrics.update(channel_std(image))
        # Data loss.
        preprocessing_metrics.update(data_loss(image))
        # Sharpness.
        preprocessing_metrics.update(sharpness(image, reduce=self.__sharpness_reduce))
        # Resize image and mask for channel quantiles.
        resized_image = resize_image(
            image, self.__channel_resize, return_arr=True, fast_resize=True
        )
        resized_mask = resize_image(
            tissue_mask,
            self.__channel_resize,
            return_arr=True,
            fast_resize=True,
        )
        # Channel quantiles.
        preprocessing_metrics.update(
            channel_quantiles(resized_image, resized_mask, self.__quantiles)
        )
        # Call custom callback.
        if self.__custom_callback is not None:
            output = self.__custom_callback(image)
            if not isinstance(output, dict):
                raise TypeError("Custom callback should return a dictionary.")
            preprocessing_metrics.update(self.__custom_callback(image))
        return preprocessing_metrics

    def __repr__(self):
        callback_name = None
        if self.__custom_callback:
            callback_name = self.__custom_callback.__name__
        return (
            "{}(channel_resize={}, quantiles={}, "
            "sharpness_reduce={}, custom_callback={})".format(
                self.__class__.__name__,
                self.__channel_resize,
                self.__quantiles,
                self.__sharpness_reduce,
                callback_name,
            )
        )


def channel_quantiles(
    image: Union[numpy.ndarray, Image.Image],
    tissue_mask: numpy.ndarray = None,
    quantiles: List[float] = DEFAULT_QUANTILES,
) -> Dict[str, List[int]]:
    """Calculate image channel quantiles which is useful in the detection of
    artifacts and shitty images. If the input image is RGB, HSV quantiles are
    also automatically calculated.

    Args:
        image: Input image.
        tissue_mask: Tissue mask for the input image. Helpful for discarding
            background in quantile evaluation. Defaults to None.
        quantiles: Quantiles to be collected. Defaults to
            [0.05, 0.1, 0.5, 0.9, 0.95].

    Returns:
        Quantiles for each image channel.

    Example:
        ```python
        import histoprep.functional as F
        from histoprep.helpers import read_image

        image = read_image("path/to/image.jpeg")
        results = F.channel_quantiles(image)
        ```
    """
    if any(x < 0 or x > 1 for x in quantiles):
        raise ValueError("Quantiles should be between 0 and 1.")
    image = _to_array(image)
    if tissue_mask is not None:
        if tissue_mask.shape != image.shape[:2]:
            raise ValueError(
                "Tissue mask shape ({}) should match image shape ({}).".format(
                    image.shape, tissue_mask.shape
                )
            )
        selection = tissue_mask == 1
        if (selection).sum() == 0:
            selection = tissue_mask >= 0
    else:
        selection = image >= 0 if image.ndim == 2 else image[..., 0] >= 0
    # Collect quantiles.
    results = []
    if image.ndim == 2:
        # Only one channel.
        bins = numpy.cumsum(numpy.bincount(image[selection].flatten(), minlength=256))
        for q in quantiles:
            results.append(int(numpy.argwhere(bins > int(q * selection.sum()))[0]))
        return {"q={}".format(q): x for q, x in zip(quantiles, results)}
    else:
        # Loop each channel.
        for img in (image, _rgb2hsv(image)):
            for c in range(3):
                bins = numpy.cumsum(
                    numpy.bincount(img[selection, c].flatten(), minlength=256)
                )
                tmp = []
                for q in quantiles:
                    tmp.append(int(numpy.argwhere(bins > int(q * selection.sum()))[0]))
                results.append(tmp)
    results_dict = {}
    for c_name, q_values in zip(
        ["red", "green", "blue", "hue", "saturation", "brightness"], results
    ):
        for q_name, q_val in zip(quantiles, q_values):
            results_dict["{}_q={}".format(c_name, q_name)] = q_val
    return results_dict


def sharpness(
    image: Union[numpy.ndarray, Image.Image],
    reduce: Union[str, List[str]] = "max",
) -> Dict[str, float]:
    """
    The method takes five crops from the image and calculates the standard
    deviation of the crops after a Laplace transformation. These values then
    reduced with the selected method.

    Args:
        image: Input image.
        reduction: Reduction method(s) for the Laplacian variance values for
            each crop. Defaults to "max".
    Returns:
        Reduced laplacian standard deviation of the crops.

    Example:
        ```python
        import histoprep.functional as F
        from histoprep.helpers import read_image

        image = read_image("path/to/image.jpeg")
        sharpness = F.sharpness(image, "median")
        ```
    """
    image = _to_array(image)
    laplacian = _to_laplacian(image)
    # Collect stds (tirsk).
    values = []
    for crop in _five_crop(laplacian):
        values.append(crop.std())
    # Then reduction(s).
    if isinstance(reduce, str):
        reduce = [reduce]
    results = dict(
        zip(
            ["sharpness_" + method for method in reduce],
            [_reduce_values(values, method) for method in reduce],
        )
    )
    return results


def data_loss(
    image: Union[numpy.ndarray, Image.Image],
) -> Dict[str, float]:
    """
    Calculates the percentage of completely white and black pixels.

    Args:
        image: Input image.

    Returns:
        Percentages of completely black (0) and white (255) pixels.

    Example:
        ```python
        import histoprep.functional as F
        from histoprep.helpers import read_image

        image = read_image("path/to/image.jpeg")
        black_pixels, white_pixels = F.data_loss(image).values()
        ```
    """
    image = _to_array(image)
    if image.ndim != 2:
        image = _to_grayscale(image)
    return {
        "black_pixels": (image == 0).sum() / image.size,
        "white_pixels": (image == 255).sum() / image.size,
    }


def channel_mean(image: Union[numpy.ndarray, Image.Image]) -> Dict[str, float]:
    """Calculate standard deviation in each image channel.

    If the image is grayscale, only the total standard deviation is returned.
    Else both RGB and HSV channel standard deviations are evaluated.

    Args:
        image: Input image.

    Returns:
        Channel standard deviations.

    Example:
        ```python
        import histoprep.functional as F
        from histoprep.helpers import read_image

        image = read_image("path/to/image.jpeg")
        channe_mean = F.channel_mean(image)
        ```
    """
    image = _to_array(image)
    if image.ndim == 2:
        return {"gray_mean": image.mean().round(3)}
    # Start with gray.
    results = {"gray_mean": _to_grayscale(image).mean().round(3)}
    # Collect both RGB and HSV stds.
    for key, val in zip(
        ["red", "green", "blue", "hue", "saturation", "brightness"],
        (*__channel_mean(image), *__channel_mean(_rgb2hsv(image))),
    ):
        results["{}_mean".format(key)] = val
    return results


def channel_std(image: Union[numpy.ndarray, Image.Image]) -> Dict[str, float]:
    """Calculate standard deviation in each image channel.

    If the image is grayscale, only the total standard deviation is returned.
    Else both RGB and HSV channel standard deviations are evaluated.

    Args:
        image: Input image.

    Returns:
        Channel standard deviations.

    Example:
        ```python
        import histoprep.functional as F
        from histoprep.helpers import read_image

        image = read_image("path/to/image.jpeg")
        channel_std = F.channel_std(image)
        ```
    """
    image = _to_array(image)
    if image.ndim == 2:
        return {"gray_std": image.std().round(3)}
    # Start with gray.
    results = {"gray_std": _to_grayscale(image).std().round(3)}
    # Collect both RGB and HSV stds.
    for key, val in zip(
        ["red", "green", "blue", "hue", "saturation", "brightness"],
        (*__channel_std(image), *__channel_std(_rgb2hsv(image))),
    ):
        results["{}_std".format(key)] = val
    return results


def __channel_mean(image: numpy.ndarray) -> Tuple[float, float, float]:
    """Calculate channel mean."""
    return [image[..., c].mean().round(3) for c in range(3)]


def __channel_std(image: numpy.ndarray) -> Tuple[float, float, float]:
    """Calculate channel std."""
    return [image[..., c].std().round(3) for c in range(3)]



import os
import shutil
from typing import Tuple, Union

# def find_files(
#     parent_dir: str, extension: Union[Tuple[str], str], depth: int = 0
# ) -> List[str]:
#     """Find all files in the ``parent_dir`` with extension.

#     Args:
#         parent_dir: Parent directory to loop.
#         extension: File extension(s).
#         depth: Depth for recursive search. Set -1 for fully recursive search.
#             Defaults to 0.

#     Returns:
#        Paths with the desired extension(s).

#     Example:
#         ```python
#         from histoprep.helpers import find_files

#         # Load spot metadata and images.
#         metadata = find_files("path/to/output_dir/", extension="csv", depth=1)
#         spots = find_files("path/to/output_dir/", extension="jpeg", depth=2)
#         ```
#     """
#     if not os.path.isdir(parent_dir):
#         raise ValueError("{} is not a directory".format(parent_dir))
#     if depth < 0 or not isinstance(depth, int):
#         raise TypeError("Depth should be a positive integer.")
#     if isinstance(extension, str):
#         if not extension.startswith("."):
#             extension = "." + extension
#     elif isinstance(extension, (tuple, list)):
#         if len(extension) > 0 and not isinstance(extension[0], str):
#             raise TypeError(
#                 "Extension should be a string not {}.".format(type(extension[0]))
#             )
#         extension = tuple(x if x.startswith(".") else "." + x for x in extension)
#     else:
#         raise ValueError(
#             "Expected extension to be a tuple or string not {}.".format(type(extension))
#         )
#     return _list_files(parent_dir, extension, 0, depth)


# def _list_files(dir_path: str, extensions: tuple, current_depth: int, max_depth: int):
#     """Helper function to list files recursively."""
#     file_paths = []
#     for f in os.scandir(dir_path):
#         if f.is_file() and f.name.endswith(extensions):
#             file_paths.append(f.path)
#         if f.is_dir() and (max_depth == -1 or current_depth + 1 <= max_depth):
#             file_paths += _list_files(f.path, extensions, current_depth + 1, max_depth)
#     return file_paths


def remove_extension(path: str) -> str:
    """Return filename with the extension removed."""
    filename = os.path.basename(path)
    if "." in filename:
        filename = ".".join(filename.split(".")[:-1])
    return os.path.join(os.path.dirname(path), filename)


def get_extension(path: str) -> str:
    """Return file extension."""
    filename = os.path.basename(path)
    if "." in filename:
        return filename.split(".")[-1]
    else:
        return None


def remove_directory(dir_path: str) -> None:
    """Remove directory and all files.

    Args:
        dir_path: Directory to be removed.

    Example:
        ```python
        from histoprep.helpers import remove_dir

        remove_dir("/shitty/slide/output_dir")
        ```
    """
    if not os.path.exists(dir_path):
        raise FileNotFoundError("Path %s does not exist." % dir_path)
    elif os.path.isfile(dir_path):
        raise ValueError("Path {} is not a directory.".format(dir_path))
    shutil.rmtree(dir_path)


def remove(path: str) -> None:
    if os.path.exists(path):
        os.remove(path)



import functools
import logging
import multiprocessing
import os
from typing import Any, Callable, Iterable, Iterator, Tuple

__all__ = ["multiprocess_loop"]


def multiprocess_loop(
    func: Callable,
    iterable: Iterable,
    num_workers: int = None,
    initializer: Callable = None,
    initializer_args: Tuple[Any] = (),
    use_imap: bool = True,
    **kwargs,
) -> Iterator[Iterable]:
    """Maps function and iteration with multiple workers and yields the outputs.

    Args:
        func: Function to be called on each list item.
        iterable: Iterable passed to the function.
        num_workers: Number of worker processes. If None, set to the number of
            CPU cores. Defaults to None.
        initializer: Initializer function for each worker process. Defaults to
            None.
        initializer_args:: Arguments for the `initializer` function. Defaults to
            ().
        use_imap: Uses imap instead of map. Imap returns results in the same
            order but is slightly slower to start up. Defaults to True.
        **kwargs: Passed to the `func`.

    Return:
        Iterable of function outputs.

    Example:
        ```python
        from histoprep.helpers import multiprocess_loop, read_image, progress_bar

        tiles = []
        for tile in multiprocess_loop(
            func=read_image,
            iterable=paths,
            num_workers=20,
            return_arr=True # <- This is a keyword argument for read_image!
        ):
            tiles.append(tile)
        ```
    """
    if is_lambda(func):
        raise AttributeError("Lambda functions cannot be pickled for multiprocessing!")
    if is_local(func):
        raise AttributeError(
            "Locally defined functions cannot be pickled for multiprocessing!"
        )
    if not isinstance(iterable, Iterable):
        raise TypeError(
            "Passed `iterable` ({}) is not iterable.".format(type(iterable))
        )
    # Define processes.
    if num_workers is None:
        num_workers = os.cpu_count()
        logging.debug("Setting num_workers to {}.".format(num_workers))
    elif num_workers < 0:
        raise ValueError(
            "Number of workers should be positive (not {}).".format(num_workers)
        )
    # Wrap function.
    func = functools.partial(func, **kwargs)
    # Start pool.
    if num_workers > 1:
        with multiprocessing.Pool(
            processes=num_workers,
            initializer=initializer,
            initargs=initializer_args,
        ) as pool:
            if use_imap:
                for result in pool.imap(func, iterable):
                    yield result
            else:
                for result in pool.map(func, iterable):
                    yield result
    else:
        # Yield with the main process.
        for x in iterable:
            yield func(x)


def is_lambda(func):
    """Test if a function is a lambda function."""
    return hasattr(func, "__qualname__") and "<lambda>" in func.__qualname__


def is_local(func):
    """Test if a function is a locally defined function."""
    return hasattr(func, "__qualname__") and "<locals>" in func.__qualname__



import itertools
import logging
from typing import List, Tuple, Union

import numpy

__all__ = ["tile_coordinates", "filter_coordinates"]


def tile_coordinates(
    dimensions: Tuple[int, int],
    width: int,
    height: int = None,
    overlap: float = 0.0,
) -> List[Tuple[int, int, int, int]]:
    """Extract a list of tile coordinates based on image dimensions.

    Args:
        dimensions: Image dimensions (height, width).
        width: Width of a tile.
        height: Height of a tile. If None, will be set to width. Defaults to None.
        overlap: Overlap between neighbouring tiles. Defaults to 0.0.

    Returns:
        Tile coordinates in XYWH format.

    Example:
        ```python
        import histoprep.functional as F
        from histoprep.helpers import read_image

        # Read image and extract tile coordinates.
        image = read_image("path/to/image.jpeg")
        coordinates = F.tile_coordinates(
            dimensions=image.size,
            width=512,
            overlap=0.25,
        )
        ```
    """
    if not isinstance(dimensions, tuple) or not isinstance(dimensions[0], int):
        raise TypeError("Dimensions should be a tuple of integers.")
    elif len(dimensions) != 2:
        raise ValueError(
            "Dimensions should contain 2 values for height and width, "
            "not {}.".format(len(dimensions))
        )
    if not 0 <= overlap < 1.0:
        raise ValueError("Overlap should be in range [0, 1).")
    if height is None:
        height = width
    if not (isinstance(width, int) and isinstance(height, int)):
        raise TypeError("Height and width should be integers.")
    if height <= 0 or width <= 0:
        raise ValueError("Height and width should be over 0.")
    if height > dimensions[0] or width > dimensions[1]:
        raise ValueError("Tile height or width is larger than image dimensions.")
    # Collect y coords.
    y = [0]
    overlap_y = int(height * overlap)
    while y[-1] < dimensions[0]:
        y.append(y[-1] + height - overlap_y)
    y = y[:-1]
    # Collect x coords.
    x = [0]
    overlap_x = int(width * overlap)
    while x[-1] < dimensions[1]:
        x.append(x[-1] + width - overlap_x)
    x = x[:-1]
    # Take product.
    coordinates = list(itertools.product(x, y))
    # Add width and height.
    coordinates = [xy + (width, height) for xy in coordinates]
    return coordinates


def filter_coordinates(
    coordinates: List[Tuple[int, int, int]],
    tissue_mask: numpy.ndarray,
    max_background: float = 0.95,
    downsample: Union[float, Tuple[float, float]] = 1.0,
) -> List[Tuple[int, int, int, int, float]]:
    """Filter a list of coordinates based on the amount of background.

    Args:
        coordinates: List of coordinates in XYWH format.
        mask: Tissue mask.
        max_background: Maximum amount of background in tile.  Defaults to 0.95.
        downsample: Downsample of the tissue mask. Defaults to 1.

    Returns:
        Filtered list of coordinates.

    Example:
        ```python
        import histoprep.functional as F
        from histoprep.helpers import read_image

        # Read image and extract tile coordinates.
        image = read_image("path/to/image.jpeg")
        coordinates = F.tile_coordinates(
            dimensions=image.size,
            width=512,
            overlap=0.25,
        )
        # Detect tissue.
        tissue_mask = F.detect_tissue(image)
        # Filter coordinates based on the amount of background.
        filtered_coordinates = F.filter_coordinates(
            coordinates=coordinates,
            tissue_mask=tissue_mask,
            max_background=0.9,
        )
        ```
    """
    if not 0 <= max_background <= 1:
        raise ValueError("Maximum baground should be in range [0,1].")
    if not isinstance(coordinates, list):
        raise TypeError("Coordinates should be a list of tuples (X, Y, W, H).")
    elif len(coordinates) == 0:
        logging.debug("Passed empty list of coordinates to filter_coordinates.")
        return coordinates
    if (
        not isinstance(coordinates[0], tuple)
        or not isinstance(coordinates[0][0], int)
        or len(coordinates[0]) != 4
    ):
        raise ValueError("Coordinates should be a list of tuples (X, Y, W, H).")
    filtered = []
    if not isinstance(downsample, (tuple, list)):
        downsample = (downsample, downsample)
    for x, y, w, h in coordinates:
        x_d = round(x / downsample[1])
        w_d = round(w / downsample[1])
        y_d = round(y / downsample[0])
        h_d = round(w / downsample[0])
        tile_mask = tissue_mask[y_d : y_d + h_d, x_d : x_d + w_d]
        if tile_mask.size > 0:
            background_percentage = (1 - tile_mask).sum() / tile_mask.size
            if background_percentage <= max_background:
                filtered.append((x, y, w, h))
    return filtered



import os

import histoprep.functional as F
import pytest
from histoprep.helpers import read_image

from paths import DATA_PATH

IMAGE = read_image(os.path.join(DATA_PATH, "tile.jpeg"))
__, MASK = F.detect_tissue(IMAGE)


def test_tile_coordinates():
    coordinates = F.tile_coordinates(
        dimensions=(256, 256), width=32, overlap=0.25
    )
    assert len(coordinates[0]) == 4
    assert coordinates[0] == (0, 0, 32, 32)
    assert len(coordinates) == 121
    coordinates = F.tile_coordinates(
        dimensions=(256, 256), width=32, height=64, overlap=0.25
    )
    assert len(coordinates[0]) == 4
    assert coordinates[0] == (0, 0, 32, 64)
    assert len(coordinates) == 66


def test_tile_coordinates_failure():
    with pytest.raises(TypeError):
        F.tile_coordinates(dimensions=(256, 256), width=1.0, overlap=0.25)
    with pytest.raises(ValueError):
        F.tile_coordinates(dimensions=(256, 256), width=300, overlap=0.25)
    with pytest.raises(ValueError):
        F.tile_coordinates(dimensions=(256, 256), width=0, overlap=0.25)
    with pytest.raises(ValueError):
        F.tile_coordinates(
            dimensions=(256, 256), width=1, height=0, overlap=0.25
        )
    with pytest.raises(ValueError):
        F.tile_coordinates(dimensions=(256, 256), width=32, overlap=-1)
    with pytest.raises(ValueError):
        F.tile_coordinates(dimensions=(256, 256), width=32, overlap=1)
    with pytest.raises(ValueError):
        F.tile_coordinates(dimensions=(256, 256), width=32, overlap=1.2423)


def test_filter_tile_coordinates():
    coords = F.tile_coordinates(dimensions=MASK.shape, width=32, overlap=0.25)
    filtered_coordinates = F.filter_coordinates(
        coords, MASK, max_background=0.5
    )
    assert len(filtered_coordinates[0]) == 4  # (X, Y, W, H)
    assert len(F.filter_coordinates([], MASK, max_background=0.5)) == 0
    assert len(F.filter_coordinates(coords, MASK, max_background=1.0)) == len(
        coords
    )


def test_filter_tile_coordinates_failure():
    coords = F.tile_coordinates(dimensions=MASK.shape, width=32, overlap=0.25)
    with pytest.raises(TypeError):
        F.filter_coordinates(-1, MASK, max_background=0.5)
    with pytest.raises(ValueError):
        F.filter_coordinates([0], MASK, max_background=0.5)
    with pytest.raises(ValueError):
        F.filter_coordinates([(0, 4235)], MASK, max_background=0.5)
    with pytest.raises(ValueError):
        F.filter_coordinates(coords, MASK, max_background=-1)
    with pytest.raises(ValueError):
        F.filter_coordinates(coords, MASK, max_background=1.2)



import logging
from typing import List, Tuple

import cv2
import numpy
from sklearn.cluster import AgglomerativeClustering
from sklearn.metrics import silhouette_score

__all__ = ["dearray"]


def dearray(
    tissue_mask: numpy.ndarray,
    kernel_size: int = 5,
    iterations: int = 3,
    min_area: float = 0.1,
    max_area: float = 3.0,
) -> Tuple[numpy.ndarray, Tuple[int, int, int, int], List[int]]:
    """Detects TMA spots from an image and gives each spot a number starting
    from the top-left corner.

    Args:
        tissue_mask: Tissue mask.
        kernel_size: Kernel size for dilate. Defaults to 5.
        iterations: Dilate iterations. Defaults to 3.
        min_area: Minimum area for a spot. Defaults to 0.1 and calculated with:
            `median(areas) * min_area`
        max_area_ Maximum area for a spot. Similar to `min_area`. Defaults to 3.0.
    """
    # Dilate.
    kernel = numpy.ones((kernel_size, kernel_size), numpy.uint8)
    spot_mask = cv2.dilate(tissue_mask, kernel, iterations=iterations)
    # Detect contours and get their areas.
    contours, __ = cv2.findContours(spot_mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)
    areas = numpy.array([cv2.contourArea(cnt) for cnt in contours])
    # Define min and max values.
    min_area = numpy.median(areas) * min_area
    max_area = numpy.median(areas) * max_area
    # Initialize new mask.
    new_mask = numpy.zeros_like(spot_mask)
    bboxes = []
    centroids = []
    for cnt in contours:
        area = cv2.contourArea(cnt)
        # Select only contours that fit into area range.
        if not min_area < area < max_area:
            continue
        # Get bounding box.
        bbox = cv2.boundingRect(cnt)
        # Get centroid.
        moments = cv2.moments(cnt)
        centroid = (
            int(moments["m10"] / moments["m00"]),
            int(moments["m01"] / moments["m00"]),
        )
        # Draw to the new map.
        cv2.drawContours(new_mask, [cnt], -1, 1, -1)
        # Add bbox and centroid.
        bboxes.append(bbox)
        centroids.append(centroid)
    # Detect possible rotation of the image based on centroids.
    theta = get_theta(centroids)
    # Rotate centroids.
    centroids = rotate_coordinates(centroids, theta)
    # Detect optimal number of rows and columns.
    num_cols = get_optimal_cluster_size(centroids[:, 0])
    num_rows = get_optimal_cluster_size(centroids[:, 1])
    # Cluster each coordinate to column and row.
    col_labels = hierachial_clustering(centroids[:, 0], n_clusters=num_cols)
    row_labels = hierachial_clustering(centroids[:, 1], n_clusters=num_rows)
    # Detect cluster means.
    x_means = [centroids[col_labels == i, 0].mean() for i in range(num_cols)]
    y_means = [centroids[row_labels == i, 1].mean() for i in range(num_rows)]
    # Change label numbers to correct order (starting from top-left).
    for i in range(num_cols):
        new_label = numpy.arange(num_cols)[numpy.argsort(x_means) == i]
        col_labels[col_labels == i] = -new_label
    col_labels *= -1
    for i in range(num_rows):
        new_label = numpy.arange(num_rows)[numpy.argsort(y_means) == i]
        row_labels[row_labels == i] = -new_label
    row_labels *= -1
    # Collect numbers.
    numbers = numpy.zeros(len(centroids)).astype("str")
    i = 1
    complain = False
    for r in range(num_rows):
        for c in range(num_cols):
            idx = [x == (c, r) for x in zip(col_labels, row_labels)]
            if sum(idx) == 1:
                numbers[idx] = i
            elif sum(idx) > 1:
                complain = True
                numbers[idx] = [f"{i}_{ii}" for ii in range(sum(idx))]
            i += 1
    if complain:
        logging.warning("Some spots were assigned the same number.")
    # Return new mask, coords and numbers.
    return new_mask, bboxes, numbers


def get_theta(centroids: numpy.ndarray) -> float:
    """Detect rotation from centroid coordinates and return angle in radians."""
    # Calculate angle between each centroid.
    n = len(centroids)
    thetas = []
    for r in range(n):
        for c in range(n):
            x1, y1 = centroids[r]
            x2, y2 = centroids[c]
            thetas.append(numpy.rad2deg(numpy.arctan2(y2 - y1, x2 - x1)))
    # We want deviations from 0 so divide corrections.
    corrections = numpy.arange(0, 361, 45)
    for i, theta in enumerate(thetas):
        sign = numpy.sign(theta)
        idx = numpy.abs(numpy.abs(theta) - corrections).argmin()
        thetas[i] = theta - sign * corrections[idx]
    # Finally return most common angle.
    values, counts = numpy.unique(numpy.round(thetas), return_counts=True)
    theta = values[counts.argmax()]
    return numpy.radians(theta)


def rotate_coordinates(coords: numpy.ndarray, theta: float) -> numpy.ndarray:
    """Rotate coordinates with given theta."""
    c, s = numpy.cos(theta), numpy.sin(theta)
    R = numpy.array(((c, -s), (s, c)))
    return coords @ R


def get_optimal_cluster_size(X: numpy.ndarray) -> int:
    """Find optimal cluster size for dataset X."""
    if len(X.shape) < 2:
        X = X.reshape(-1, 1)
    sil = []
    for n in range(2, X.shape[0]):
        clust = AgglomerativeClustering(n_clusters=n, linkage="ward")
        clust.fit(X)
        sil.append(silhouette_score(X, clust.labels_))
    return numpy.argmax(sil) + 2


def hierachial_clustering(
    X: numpy.ndarray, n_clusters: int, linkage: str = "ward"
) -> numpy.ndarray:
    """Perform hierarchian clustering."""
    if len(X.shape) < 2:
        X = X.reshape(-1, 1)
    clust = AgglomerativeClustering(n_clusters=n_clusters, linkage=linkage)
    clust.fit(X)
    return clust.labels_



from ._files import remove_directory
from ._io import read_image
from ._metadata import combine_metadata, rename_paths, strip_metric_colums
from ._multiprocess import multiprocess_loop
from ._verbose import progress_bar
from ._visualize import random_tile_collage



import os

from histoprep.helpers._files import get_extension, remove_extension

DIR_PATH = os.path.dirname(__file__)


def test_remove_extension():
    for src, dst in {
        "/data.p": "/data",
        "./data.p": "./data",
        "/d.a.t.a.p": "/d.a.t.a",
        "./pickle": "./pickle",
    }.items():
        assert dst == remove_extension(src)


def test_get_extension():
    for src, dst in {
        "/data.p": "p",
        "./data.p": "p",
        "/d.a.t.a.p": "p",
        "./pickle": None,
    }.items():
        assert dst == get_extension(src)


# def test_find_files():
#     assert len(find_files(DIR_PATH, ".py", depth=0)) > 0
#     assert len(find_files(DIR_PATH, "py", depth=0)) > 0
#     assert len(find_files(DIR_PATH, "py", depth=0)) < len(
#         find_files(DIR_PATH, "py", depth=1)
#     )
#     assert len(find_files(DIR_PATH, "ekfrhjweofgiuh", depth=100)) == 0
#     assert len(find_files(DIR_PATH, "jpeg", depth=0)) == 0
#     assert len(find_files(DIR_PATH, "jpeg", depth=1)) == 1



import os

import pytest
from histoprep.functional import (
    PreprocessMetrics,
    channel_quantiles,
    data_loss,
    detect_tissue,
    sharpness,
)
from histoprep.functional._functional import _gaussian_blur, _to_grayscale
from histoprep.functional._helpers import _to_array
from histoprep.helpers import read_image

from paths import DATA_PATH

IMAGE = read_image(os.path.join(DATA_PATH, "tile.jpeg"))
__, MASK = detect_tissue(IMAGE)


def test_opencv():
    from histoprep.functional._functional import (
        _gaussian_blur,
        _rgb2hsv,
        _thresholding,
        _to_grayscale,
        _to_laplacian,
    )

    img = _to_array(IMAGE)
    gray = _to_grayscale(img)
    blur = _gaussian_blur(gray, 2.0)
    _thresholding(blur)
    _thresholding(blur, 150)
    _to_laplacian(gray)
    _rgb2hsv(img)


def test_channel_quantiles():
    arr_image = _to_array(IMAGE)
    gray_image = (_to_grayscale(arr_image) * 255).astype("uint8")
    channel_quantiles(IMAGE, MASK)
    channel_quantiles(arr_image, MASK)
    channel_quantiles(gray_image, MASK)


def test_passing_bad_quantiles_to_channel_quantiles():
    try:
        channel_quantiles(IMAGE, MASK, quantiles=[-1, 0])
    except ValueError:
        pass


def test_data_loss_function():
    arr_image = _to_array(IMAGE)
    gray_image = (_to_grayscale(arr_image) * 255).astype("uint8")
    data_loss(IMAGE)
    data_loss(arr_image)
    data_loss(gray_image)


def test_data_loss_outputs():
    # No loss.
    no_loss = _to_array(IMAGE)
    assert data_loss(no_loss)["white_pixels"] == 0.0
    assert data_loss(no_loss)["black_pixels"] == 0.0
    # White loss.
    white_pixels = no_loss.copy()
    white_pixels[:64, :64] = 255
    assert data_loss(white_pixels)["white_pixels"] > 0.0
    assert data_loss(white_pixels)["black_pixels"] == 0.0
    # Black loss.
    white_pixels = no_loss.copy()
    white_pixels[:64, :64] = 0
    assert data_loss(white_pixels)["white_pixels"] == 0.0
    assert data_loss(white_pixels)["black_pixels"] > 0.0
    # Both losses.
    both_losses = no_loss.copy()
    both_losses[:64, :64] = 0
    both_losses[64:, 64:] = 255
    assert data_loss(both_losses)["white_pixels"] > 0.0
    assert data_loss(both_losses)["black_pixels"] > 0.0


def test_sharpenss_function():
    arr_image = _to_array(IMAGE)
    gray_image = (_to_grayscale(arr_image) * 255).astype("uint8")
    sharpness(IMAGE)
    sharpness(arr_image)
    sharpness(gray_image)


def test_sharpness_function_with_blurred_image():
    blur = _gaussian_blur(_to_array(IMAGE), sigma=2)
    sharpness(IMAGE)["sharpness_max"] > sharpness(blur)["sharpness_max"]


def test_PreprocessMetrics():
    # Normal usage.
    metrics = PreprocessMetrics()
    # repr
    print(metrics)
    with pytest.raises(TypeError):
        # without threshold
        metrics(IMAGE)
    res = metrics(IMAGE, 200)
    assert isinstance(res, dict)
    # Sharpness.
    metrics = PreprocessMetrics(sharpness_reduce=["max", "min", "median"])
    res = metrics(IMAGE, 200)
    assert "sharpness_max" in res
    assert "sharpness_min" in res
    assert "sharpness_median" in res
    # Custom callback
    metrics = PreprocessMetrics(custom_callback=lambda x: {"dog": "good_boi"})
    res = metrics(IMAGE, 200)
    assert res["dog"] == "good_boi"
    # Bad threshold.
    metrics = PreprocessMetrics(custom_callback=lambda x: {"dog": "good_boi"})
    assert metrics(IMAGE, 0)["background"] == 1.0



from typing import Dict, Tuple

import numpy

try:
    import openslide
except Exception as e:
    raise ImportError(
        "Could not import `openslide` due to error {}. Make sure you have "
        "OpenSlide installed (https://openslide.org/download/).".format(e)
    )

from ._base import Backend

__all__ = ["OpenSlideBackend"]

OPENSLIDE_READABLE = (
    "svs",
    "svslide",
    "tif",
    "tiff",
    "vms",
    "vmu",
    "ndpi",
    "bif",
    "scn",
    "mrxs",
)


class OpenSlideBackend(Backend):
    def __init__(self, path: str):
        self.reader = openslide.OpenSlide(path)
        # Init caches.
        self.__level_dimensions = None
        self.__level_downsamples = None

    def get_dimensions(self) -> Tuple[int, int]:
        return self.level_dimensions[0]

    def get_level_dimensions(self) -> Dict[int, Tuple[int, int]]:
        if self.__level_dimensions is None:
            level_dimensions = {}
            for level, __ in enumerate(self.reader.level_downsamples):
                # OpenSlide returns X, Y and we want Y, X
                level_dimensions[level] = self.reader.level_dimensions[level][::-1]
            self.__level_dimensions = level_dimensions
        return self.__level_dimensions

    def get_level_downsamples(self) -> Dict[int, Tuple[int, int]]:
        if self.__level_downsamples is None:
            level_downsamples = {}
            for level, (y, x) in self.level_dimensions.items():
                level_downsamples[level] = (
                    self.dimensions[0] / y,
                    self.dimensions[1] / x,
                )
            self.__level_downsamples = level_downsamples
        return self.__level_downsamples

    def get_thumbnail(self, level: int) -> numpy.ndarray:
        # Read thumbnail.
        thumbnail = self.reader.get_thumbnail(self.reader.level_dimensions[level])
        # Convert to array and select only RGB channels (no alpha channel).
        return numpy.array(thumbnail)[..., :3]

    def read_region(self, XYWH: Tuple[int, int, int, int], level: int) -> numpy.ndarray:
        # Unpack.
        x, y, w, h = XYWH
        # Read region
        tile = self.reader.read_region(location=(x, y), level=level, size=(w, h))
        # Convert to array and select only RGB channels (no alpha channel).
        return numpy.array(tile)[..., :3]

    def __repr__(self):
        return "OPENSLIDE"



from typing import Dict, Tuple

import numpy


class Backend:
    def __init__(self):
        pass

    @property
    def dimensions(self) -> Tuple[int, int]:
        return self.get_dimensions()

    @property
    def level_downsamples(self) -> Dict[int, Tuple[float, float]]:
        return self.get_level_downsamples()

    @property
    def level_dimensions(self) -> Dict[int, Tuple[int, int]]:
        return self.get_level_dimensions()

    def get_dimensions(self) -> Tuple[int, int]:
        raise NotImplementedError()

    def get_level_downsamples(self) -> Dict[int, Tuple[int, int]]:
        raise NotImplementedError()

    def get_level_dimensions(self) -> Dict[int, Tuple[int, int]]:
        raise NotImplementedError()

    def get_thumbnail(self, level: int) -> numpy.ndarray:
        raise NotImplementedError()

    def read_region(self, XYWH: Tuple[int, int, int, int], level: int) -> numpy.ndarray:
        raise NotImplementedError()



import time
from contextlib import contextmanager


@contextmanager
def timeit() -> float:
    """Use for easy speed tests."""
    start = time.perf_counter()
    yield lambda: time.perf_counter() - start


def format_seconds(n: float) -> str:
    """Format seconds into pretty string format."""
    if n < 0:
        raise ValueError("n should be positive.")
    years = int(n // (365.25 * 24 * 3600))
    n %= 365.25 * 24 * 3600
    days = int(n // (24 * 3600))
    n %= 24 * 3600
    hours = int(n // 3600)
    n %= 3600
    minutes = int(n // 60)
    n %= 60
    seconds = n
    if years > 0:
        strtime = f"{years}y {days}d {(hours)}h:{minutes}m:{int(seconds)}s"
    elif days > 0:
        strtime = f"{days}d {(hours)}h:{minutes}m:{int(seconds)}s"
    elif hours > 0:
        strtime = f"{(hours)}h:{minutes}m:{int(seconds)}s"
    elif minutes > 0:
        strtime = f"{minutes}m:{int(seconds)}s"
    else:
        strtime = f"{seconds:.3g}s"
    return strtime


def strtime():
    """Returns date and timestamp as a string."""
    return time.strftime("%h_%d_%H:%M")


def get_ETC(times: list, num_left: int):
    """Calculates ETC based on mean times and number of iters left."""
    if len(times) == 0:
        return ""
    else:
        seconds = sum(times) / len(times)
        return f"ETC: {format_seconds(seconds * num_left)}"



import os
import shutil

import histoprep
import pandas
import pytest
from PIL import Image

from paths import DATA_PATH

SLIDE_1 = os.path.join(DATA_PATH, "slide_1.svs")
SLIDE_2 = os.path.join(DATA_PATH, "slide_2.mrxs")
SLIDE_3 = os.path.join(DATA_PATH, "slide_3.czi")
SLIDE_4 = os.path.join(DATA_PATH, "tile.jpeg")
OUTPUT_DIR = os.path.join(DATA_PATH, "output")


def test_cutting_tiles():
    for path, width in [
        (SLIDE_1, 1024),
        (SLIDE_2, 1024),
        (SLIDE_3, 1024),
        (SLIDE_4, 32),
    ]:
        reader = histoprep.SlideReader(path)
        # Get coords.
        coordinates = reader.get_tile_coordinates(
            width, overlap=0.1, max_background=0.1
        )[:2]
        # Save tiles.
        reader.save_tiles(OUTPUT_DIR, coordinates=coordinates, num_workers=1)
        assert isinstance(reader.tile_metadata, pandas.DataFrame)
        assert isinstance(reader.annotated_thumbnail_tiles, Image.Image)
    # Overwrite is False.
    with pytest.warns(UserWarning):
        assert (
            reader.save_tiles(
                OUTPUT_DIR, coordinates=coordinates, num_workers=1
            )
            is None
        )
    shutil.rmtree(OUTPUT_DIR)



import os
from typing import Iterable, List, Union

import numpy
from PIL import Image

from ..functional import resize_image
from ._multiprocess import multiprocess_loop

__all__ = ["read_image"]


def read_image(
    path: str, return_arr: bool = False
) -> Union[Image.Image, numpy.ndarray]:
    """Read image.

    Args:
        path: Path to image
        return_arr: Return array instead of PIL image. Defaults to False.

    Returns:
        Image from path.

    Example:
        ```python
        from histoprep.helpers import read_image

        image = read_image("path/to/image")
        arr = read_image("path/to/image", return_arr=True)
        ```
    """
    image = Image.open(path)
    if return_arr:
        return numpy.array(image)
    return image


def read_and_resize(paths: List[str], px: int) -> Iterable:
    """Helper function to load tiles with many workers."""
    for tile in multiprocess_loop(
        __load_and_resize,
        iterable=paths,
        use_imap=False,
        num_workers=min(len(paths), os.cpu_count()),
        px=px,
    ):
        yield tile


def __load_and_resize(path: str, px: int) -> numpy.ndarray:
    """Helper function to load an image and resize it immediatly."""
    if path is None:
        return numpy.zeros((px, px, 3), dtype=numpy.uint8) + 255
    return resize_image(read_image(path), (px, px), return_arr=True, fast_resize=True)



# DEPRACATED (for now at least)

# import logging
# import warnings

# import bioformats
# import javabridge
# import warnings

# JAVABRIDGE_RUNNING = False
# JAVABRIDGE_KILLED = False


# def _init_logger():
#     """Set logging level to WARN."""
#     rootLoggerName = javabridge.get_static_field(
#         "org/slf4j/Logger", "ROOT_LOGGER_NAME", "Ljava/lang/String;"
#     )
#     rootLogger = javabridge.static_call(
#         "org/slf4j/LoggerFactory",
#         "getLogger",
#         "(Ljava/lang/String;)Lorg/slf4j/Logger;",
#         rootLoggerName,
#     )
#     logLevel = javabridge.get_static_field(
#         "ch/qos/logback/classic/Level", "WARN", "Lch/qos/logback/classic/Level;"
#     )
#     javabridge.call(
#         rootLogger, "setLevel", "(Lch/qos/logback/classic/Level;)V", logLevel
#     )


# def kill_javabrige():
#     """Kills javabridge session."""
#     global JAVABRIDGE_KILLED
#     if JAVABRIDGE_RUNNING:
#         logging.debug("Killing javabridge.")
#         javabridge.kill_vm()
#         JAVABRIDGE_KILLED = True


# def start_javabridge(silent: bool = False):
#     """Starts javabridge session."""
#     global JAVABRIDGE_RUNNING
#     # Start javabridge.
#     if JAVABRIDGE_KILLED:
#         raise RuntimeError(
#             "Javabridge session has been killed and cannot be restarted. "
#             "Restart session."
#         )
#     if not JAVABRIDGE_RUNNING:
#         if not silent:
#             warnings.warn(
#                 "Initializing javabridge. Call histoprep.kill_javabridge() at "
#                 "the end of your program."
#             )
#         javabridge.start_vm(class_path=bioformats.JARS)
#         _init_logger()
#         JAVABRIDGE_RUNNING = True



import math
import random
from typing import List, Union

import numpy
import pandas
from PIL import Image

from ._io import read_and_resize

__all__ = ["random_tile_collage"]


def random_tile_collage(
    paths: Union[List[str], pandas.Series, numpy.ndarray],
    nrows: int = 16,
    ncols: int = 32,
    px: int = 32,
) -> Image.Image:
    """Plot random selection of the given paths.

    Args:
        paths: Image paths.
        nrows: Number of rows in collage. Defaults to 16.
        ncols:  Number of columns in collage. Defaults to 32.
        px: Size of each tile in collage. Defaults to 32.

    Returns:
        Collage image of random tiles.

    Example:
        ```python
        import matplotlib.pyplot as plt
        from histoprep.helpers import random_tile_collage, combine_metadata

        # Load metadata.
        metadata = combine_metadata("/output_dir/")
        data_loss_paths = metadata["black_pixels" > 0.05]["path"]
        # Plot some tiles with data loss.
        plt.imshow(random_tile_collage(data_loss_paths))
        ```
    """
    if isinstance(paths, (pandas.Series, numpy.ndarray)):
        paths = paths.tolist()
    if not isinstance(paths, list):
        raise TypeError("Expected a list of paths, not {}.".format(type(paths)))
    elif len(paths) == 0:
        return
    elif not isinstance(paths[0], str):
        raise TypeError(
            "Expected list items to be 'str', not {}.".format(type(paths[0]))
        )
    if not isinstance(ncols, int) or not isinstance(nrows, int):
        raise TypeError("Number of columns and rows should be integers.")
    if not isinstance(px, int):
        raise TypeError("Tile size should be an integer.")
    # Select random paths.
    paths = random.sample(paths, min(nrows * ncols, len(paths)))
    # Remove rows if necessary.
    nrows = math.ceil(len(paths) / ncols)
    while len(paths) < nrows * ncols:
        paths.append(None)
    # Build collage.
    row = []
    collage = []
    for tile in read_and_resize(paths, px):
        row.append(tile)
        if len(row) == ncols:
            collage.append(numpy.hstack(row))
            row = []
    return Image.fromarray(numpy.vstack(collage))



from typing import Union

import numpy
from PIL import Image

from ._functional import _gaussian_blur, _thresholding, _to_grayscale
from ._helpers import _to_array

__all__ = ["detect_tissue"]


def detect_tissue(
    image: Union[numpy.ndarray, Image.Image],
    threshold: int = None,
    multiplier: float = 1.0,
    sigma: float = 1.0,
    remove_white: bool = False,
) -> numpy.ndarray:
    """Detect tissue from image.

    Args:
        image: Input image.
        threshold: Threshold for tissue detection. If set, will detect tissue by
            global thresholding, and otherwise Otsu's method is used to find
            a threshold. Defaults to None.
        multiplier: Otsu's method is used to find an optimal threshold by
            minimizing the weighted within-class variance. This threshold is
            then multiplied with `multiplier`. Used only if `threshold` is None.
            Defaults to 1.0.
        sigma: Sigma for gaussian blurring. Defaults to 1.0.
        remove_white: Does not consider white pixels with Otsu's method. Useful
            for slide images where large areas are artificially set to white.
            Defaults to False.

    Returns:
        Binary mask with 0=background and 1=tissue.

    Example:
        ```python
        import histoprep.functional as F
        from histoprep.helpers import read_image

        image = read_image("path/to/image.jpeg")
        tissue_mask = F.detect_tissue(image)
        ```
    """
    if threshold is not None and (
        not isinstance(threshold, int) or not 0 <= threshold <= 255
    ):
        raise TypeError("Threshold should be an integer in range [0, 255]")
    if not isinstance(multiplier, float) or multiplier <= 0:
        raise TypeError("Threshold multiplier should be a float over 0.")
    if not isinstance(sigma, (float, int)) or sigma < 0:
        raise TypeError("Sigma should be a positive float.")
    # Check image and convert to array.
    image = _to_array(image)
    # To grayscale.
    gray = _to_grayscale(image)
    if sigma > 0.0:
        # Gaussian blurring.
        gray = _gaussian_blur(gray, sigma)
    # Global thresholding.
    return _thresholding(
        gray=gray,
        threshold=threshold,
        multiplier=multiplier,
        remove_white=remove_white,
    )



from ._coordinates import filter_coordinates, tile_coordinates
from ._dearray import dearray
from ._helpers import arr2pil, downsample_image, resize_image, rgb2gray, rgb2hsv
from ._preprocess import (
    PreprocessMetrics,
    channel_quantiles,
    channel_std,
    data_loss,
    sharpness,
)
from ._tissue import detect_tissue



import functools
import time

from histoprep.helpers._multiprocess import is_lambda, is_local, multiprocess_loop


def divide_by_2(x: int, round_numbers: bool):
    x = x / 2
    if round_numbers:
        x = round(x, 0)
    return x


def divide_by_zero(i):
    1 / 0


def sleeping_beaty(i):
    time.sleep(1)


def test_multiprocess_loop():
    loop_rounding = multiprocess_loop(
        divide_by_2,
        range(20),
        num_workers=1,
        round_numbers=True,
    )
    for i, output in enumerate(loop_rounding):
        assert round(i / 2, 0) == output
    loop_no_rounding = multiprocess_loop(
        divide_by_2,
        range(20),
        num_workers=1,
        round_numbers=False,
    )
    for i, output in enumerate(loop_no_rounding):
        assert i / 2 == output


def test_multiprocess_loop_lambda_function_fail():
    assert is_lambda(lambda x: x)
    assert not is_lambda(functools.partial(divide_by_2, {"round_numbers": True}))
    assert not is_lambda(sleeping_beaty)


def test_multiprocess_loop_local_function_fail():
    def local_func(x):
        pass

    assert is_local(local_func)
    assert not is_local(functools.partial(divide_by_2, {"round_numbers": True}))
    assert not is_local(sleeping_beaty)



from typing import Dict, Tuple

import numpy
from PIL import Image

from ...functional import resize_image
from ._base import Backend

__all__ = ["PillowBackend"]

PILLOW_READABLE = ("jpeg", "jpg", "png")


class PillowBackend(Backend):
    def __init__(self, path: str):
        # Read image file.
        self.__level_images = {0: Image.open(path)}
        # Add downsamples.
        level = 1
        self.__level_dimensions = {0: self.dimensions}
        self.__level_downsamples = {0: (1.0, 1.0)}
        while True:
            if max(self.dimensions) // 2**level < 128:
                break
            h, w = tuple(x // 2**level for x in self.dimensions)
            self.__level_dimensions[level] = (h, w)
            self.__level_downsamples[level] = (
                self.dimensions[0] / h,
                self.dimensions[0] / w,
            )
            level += 1

    def get_dimensions(self) -> Dict[int, Tuple[int, int]]:
        return self.__level_images[0].size[::-1]

    def get_level_dimensions(self) -> Dict[int, Tuple[int, int]]:
        return self.__level_dimensions

    def get_level_downsamples(self) -> Dict[int, Tuple[int, int]]:
        return self.__level_downsamples

    def get_thumbnail(self, level: int) -> numpy.ndarray:
        if level not in self.__level_images.keys():
            self.__level_images[level] = resize_image(
                self.__level_images[0],
                self.__level_dimensions[level],
                fast_resize=True,
            )
        return numpy.array(self.__level_images[level])

    def read_region(self, XYWH: Tuple[int, int, int, int], level: int) -> numpy.ndarray:
        # Unpack.
        x, y, w, h = XYWH
        if level not in self.__level_images.keys():
            self.__level_images[level] = resize_image(
                self.__level_images[0],
                self.__level_dimensions[level],
                fast_resize=True,
            )
        # Read region
        return numpy.array(self.__level_images[level].crop((y, x, y + h, x + w)))

    def __repr__(self):
        return "PILLOW"



import math
import warnings
from typing import List, Tuple

import matplotlib.pyplot as plt
import numpy
import pandas
from PIL import Image

from ..helpers._io import read_and_resize

FIGURE_WIDTH, FIGURE_HEIGHT = plt.rcParams["figure.figsize"]

__all__ = ["OutlierVisualizer"]


class OutlierVisualizer:
    def __init__(self, metadata: pandas.DataFrame):
        """Visualise preprocessing metrics to identify outliers.

        Args:
            metadata: Metadata with preprocessing metrics.

        Example:
            ```python
            import histoprep
            # Initialize class.
            visualise = histoprep.OutlierVisualizer(metadata)
            # Lets visualise tiles with data loss...
            visualise.plot_histogram_with_examples(column="black_pixels")
            # ... or with too much background ...
            visualise.plot_histogram_with_examples(column="background")
            # ... or if the colour channel variances picked up any outlies.
            visualise.plot_rgb_std(log_scale=True)
            visualise.plot_hsv_std(log_scale=True)
            ```
        """
        self.__metadata = metadata
        self.__all_paths = metadata["path"].to_numpy()
        if len(self.__metadata) > 1e6:
            warnings.warn(
                "Plotting functions may take a while due to size of the data."
            )

    def plot_rgb_mean(
        self,
        bins: int = 50,
        log_scale: bool = False,
        figsize: Tuple[float, float] = (FIGURE_WIDTH * 2, FIGURE_HEIGHT),
    ):
        """Plot the standard deviation of RGB (and gray) channels for each tile.

        Args:
            bins: Number of bins. Defaults to 50.
            log_scale: Set y-axis to logarithmic scale. Defaults to False.
            figsize: Figure size. Defaults to (default_w * 2, default_h).
        """
        self.__histogram(
            columns=["gray_mean", "red_mean", "green_mean", "blue_mean"],
            titles=["Gray", "Red", "Green", "Blue"],
            log_scale=log_scale,
            subplot_kwargs=dict(nrows=1, ncols=4, figsize=figsize, sharey=True),
            hist_kwargs=dict(bins=bins),
        )
        plt.suptitle("Standard deviation color channels for each tile.")
        plt.tight_layout()

    def plot_rgb_std(
        self,
        bins: int = 50,
        log_scale: bool = False,
        figsize: Tuple[float, float] = (FIGURE_WIDTH * 2, FIGURE_HEIGHT),
    ):
        """Plot the standard deviation of RGB (and gray) channels for each tile.

        Args:
            bins: Number of bins. Defaults to 50.
            log_scale: Set y-axis to logarithmic scale. Defaults to False.
            figsize: Figure size. Defaults to (default_w * 2, default_h).
        """
        self.__histogram(
            columns=["gray_std", "red_std", "green_std", "blue_std"],
            titles=["Gray", "Red", "Green", "Blue"],
            log_scale=log_scale,
            subplot_kwargs=dict(nrows=1, ncols=4, figsize=figsize, sharey=True),
            hist_kwargs=dict(bins=bins),
        )
        plt.suptitle("Standard deviation color channels for each tile.")
        plt.tight_layout()

    def plot_hsv_mean(
        self,
        bins: int = 50,
        log_scale: bool = False,
        figsize: Tuple[float, float] = (FIGURE_WIDTH * 1.5, FIGURE_HEIGHT),
    ):
        """Plot the standard deviation of HSV channels for each tile.

        Args:
            bins: Number of bins. Defaults to 50.
            log_scale: Set y-axis to logarithmic scale. Defaults to False.
            figsize: Figure size. Defaults to (default_w * 1.5, default_h).
        """
        self.__histogram(
            columns=["hue_mean", "saturation_mean", "brightness_mean"],
            titles=["Hue", "Saturation", "Brightness"],
            log_scale=log_scale,
            subplot_kwargs=dict(nrows=1, ncols=3, figsize=figsize, sharey=True),
            hist_kwargs=dict(bins=bins),
        )
        plt.suptitle("Standard deviation of HSV channels for each tiles.")
        plt.tight_layout()

    def plot_hsv_std(
        self,
        bins: int = 50,
        log_scale: bool = False,
        figsize: Tuple[float, float] = (FIGURE_WIDTH * 1.5, FIGURE_HEIGHT),
    ):
        """Plot the standard deviation of HSV channels for each tile.

        Args:
            bins: Number of bins. Defaults to 50.
            log_scale: Set y-axis to logarithmic scale. Defaults to False.
            figsize: Figure size. Defaults to (default_w * 1.5, default_h).
        """
        self.__histogram(
            columns=["hue_std", "saturation_std", "brightness_std"],
            titles=["Hue", "Saturation", "Brightness"],
            log_scale=log_scale,
            subplot_kwargs=dict(nrows=1, ncols=3, figsize=figsize, sharey=True),
            hist_kwargs=dict(bins=bins),
        )
        plt.suptitle("Standard deviation of HSV channels for each tiles.")
        plt.tight_layout()

    @property
    def __quantile_columns(self):
        columns = []
        for col in self.__metadata.columns:
            if "_q=" in col:
                columns.append(col)
        return columns

    def plot_rgb_quantiles(
        self,
        bins: int = 50,
        log_scale: bool = False,
        figsize: Tuple[float, float] = (FIGURE_WIDTH * 2, FIGURE_HEIGHT * 2),
    ):
        """Plot the quantile values of RGB channels for each tile.

        Args:
            bins: Number of bins. Defaults to 50.
            log_scale: Set y-axis to logarithmic scale. Defaults to False.
            figsize: Figure size. Defaults to (default_w * 2, default_h * 2).
        """
        channels = ["red", "green", "blue"]
        columns = []
        titles = []
        for col in self.__quantile_columns:
            if any(x in col for x in channels):
                columns.append(col)
                titles.append("Quantile={}".format(col.split("_")[-1][2:]))
        ncols = sum("red" in x for x in columns)
        nrows = math.ceil(len(columns) / ncols)
        axes = self.__histogram(
            columns=columns,
            titles=titles,
            log_scale=log_scale,
            subplot_kwargs=dict(
                nrows=nrows,
                ncols=ncols,
                figsize=figsize,
                sharey=True,
                sharex=True,
            ),
            hist_kwargs=dict(bins=bins),
        )
        for idx in range(ncols * nrows):
            if idx % ncols == 0:
                axes[idx].set_ylabel(channels.pop().capitalize())
            axes[idx].set_xlim(0, 255)
        plt.suptitle("Quantile values of RGB channels for each tile.")
        plt.tight_layout()

    def plot_hsv_quantiles(
        self,
        bins: int = 50,
        log_scale: bool = False,
        figsize: Tuple[float, float] = (FIGURE_WIDTH * 2, FIGURE_HEIGHT * 2),
    ):
        """Plot the quantile values of RGB channels for each tile.

        Args:
            bins: Number of bins. Defaults to 50.
            log_scale: Set y-axis to logarithmic scale. Defaults to False.
            figsize: Figure size. Defaults to (default_w * 2, default_h * 2).
        """
        channels = ["hue", "saturation", "brightness"]
        columns = []
        titles = []
        for col in self.__quantile_columns:
            if any(x in col for x in channels):
                columns.append(col)
                titles.append("Quantile={}".format(col.split("_")[-1][2:]))
        ncols = sum("hue" in x for x in columns)
        nrows = math.ceil(len(columns) / ncols)
        axes = self.__histogram(
            columns=columns,
            titles=titles,
            log_scale=log_scale,
            subplot_kwargs=dict(
                nrows=nrows,
                ncols=ncols,
                figsize=figsize,
                sharey=True,
                sharex=True,
            ),
            hist_kwargs=dict(bins=bins),
        )
        for idx in range(ncols * nrows):
            if idx % ncols == 0:
                axes[idx].set_ylabel(channels.pop().capitalize())
            axes[idx].set_xlim(0, 255)
        plt.suptitle("Quantile values of HSV channels for each tile.")
        plt.tight_layout()

    def plot_histogram_with_examples(
        self,
        column: str,
        bins: int = 30,
        num_examples: int = 16,
        log_scale: bool = True,
        figsize: Tuple[float, float] = (FIGURE_WIDTH * 2, FIGURE_HEIGHT * 2),
    ):
        """Plot a histogram of column values, with example tiles from each bin.

        Args:
            column: Name of the column to plot.
            bins: Number of bins. Defaults to 30.
            num_examples: Number of example images per bin. Defaults to 16.
            log_scale: Set y-axis to logarithmic scale. Defaults to True.
            figsize: Figure size. Defaults to (default_w * 2, default_h * 2).
        """
        if column not in self.__metadata.columns:
            raise ValueError("Column {} not in metadata columns.".format(column))
        # Define figure.
        shape = (4 if num_examples <= 8 else 5, bins)
        fig = plt.figure(figsize=figsize)
        ax_hist = plt.subplot2grid(shape, (0, 0), colspan=bins)
        axes = []
        for i in range(bins):
            axes.append(
                plt.subplot2grid(shape, (1, i), colspan=1, rowspan=shape[0] - 1)
            )
        # Plot histogram.
        self.__metadata[column].hist(bins=bins, ec="black", ax=ax_hist)
        ax_hist.set_xlim(self.__metadata[column].min(), self.__metadata[column].max())
        if log_scale:
            ax_hist.set_yscale("log")
        # Load example images from each bin.
        bin_numbers = pandas.cut(
            self.__metadata[column], bins, labels=range(bins)
        ).to_numpy()
        paths = []
        for i in range(bins):
            bin_paths = self.__all_paths[bin_numbers == i]
            if len(bin_paths) == 0:
                paths += [None] * num_examples
            elif len(bin_paths) < num_examples:
                paths += bin_paths.tolist()
                paths += [None] * (num_examples - len(bin_paths))
            else:
                paths += numpy.random.choice(bin_paths, num_examples).tolist()
        # Read and plot examples.
        idx = 0
        tmp = []
        for tile in read_and_resize(paths, px=128):
            tmp.append(tile)
            if len(tmp) == num_examples:
                collage = Image.fromarray(numpy.vstack(tmp))
                axes[idx].imshow(collage)
                axes[idx].axis("off")
                idx += 1
                tmp = []
        plt.suptitle("Histogram of values from column: {}".format(column))
        plt.tight_layout(w_pad=0, h_pad=0.2)

    def __histogram(
        self,
        columns: List[str],
        titles: List[str],
        log_scale: bool,
        subplot_kwargs: dict,
        hist_kwargs: dict,
    ):
        """Helper function to plot histograms."""
        fig, axes = plt.subplots(**subplot_kwargs)
        axes = axes.ravel()
        for i, (column, title) in enumerate(zip(columns, titles)):
            self.__metadata[column].hist(ax=axes[i], **hist_kwargs)
            axes[i].set_title(title)
            if log_scale:
                axes[i].set_yscale("log")
        return axes



import logging
import os
import shutil
import time
from typing import Callable, Dict, List, Tuple, Union

import cv2
import numpy
from PIL import Image, ImageDraw

from ..helpers import remove_directory


def check_region(XYWH: tuple, dimensions: tuple) -> Tuple[bool, Tuple[int]]:
    """Checks if region is out of bounds and calculates padding."""
    x, y, w, h = XYWH
    # Check dimensions.
    out_of_bounds = False
    padding = None
    if y > dimensions[0] or x > dimensions[1]:
        # Starting point overbound at either axis --> No way to fix.
        out_of_bounds = True
    elif y + h > dimensions[0] or x + w > dimensions[1]:
        # Height or width causes overbound --> read what we can and pad.
        out_of_bounds = True
        # Get allowed h and w --> padding..
        padding = (
            min(dimensions[0] - y, h),
            min(dimensions[1] - x, w),
        )
    return out_of_bounds, padding


def copy_to_scratch(src: str, dst: str, verbose: bool = True):
    """Copy slide to scratch directory."""
    if not os.path.exists(dst):
        # Copy only if the path doesn't exist.
        shutil.copyfile(src, dst, follow_symlinks=True)
    # See if there is a separate data folder.
    if src.endswith(".mrxs"):
        logging.debug("Copying the mrxs data files.")
        src_dir = src.replace(".mrxs", "")
        dst_dir = dst.replace(".mrxs", "")
        os.makedirs(dst_dir, exist_ok=True)
        for f in os.scandir(src_dir):
            shutil.copyfile(f.path, os.path.join(dst_dir, f.name))
    # Return new path
    return dst


def find_level(preferred: int, maximum: int, level_dimensions: tuple) -> int:
    """Helper function to find a thumbnail level."""
    level = None
    for _level, dimensions in level_dimensions.items():
        if max(dimensions) <= maximum:
            level = _level
            if max(dimensions) <= preferred:
                break
    if level is None:
        raise ValueError(
            "No level with all dimensions less than {}. Dimensions for each"
            " level are:\n{}".format(maximum, level_dimensions)
        )
    return level


def check_level(level: int, level_dimensions: tuple) -> None:
    # Check level.
    if level < 0:
        # Indexing from the end.
        level += len(level_dimensions)
    if level not in level_dimensions:
        raise ValueError(
            "Level {} not available. Please choose from {}.".format(
                level, list(level_dimensions.keys())
            )
        )


def prepare_paths_and_directories(
    output_dir: str,
    slide_name: str,
    overwrite: bool,
) -> Dict[str, str]:
    """Prepare paths and directories."""
    # Define slide output directory.
    slide_dir = os.path.join(output_dir, slide_name)
    # Check both directory paths.
    for dir_path in [output_dir, slide_dir]:
        if os.path.exists(dir_path) and not os.path.isdir(dir_path):
            raise NotADirectoryError(
                "Slide directory {} exists and but isn't a directory.".format(dir_path)
            )
    # Prepare rest of the paths.
    tile_directory = os.path.join(slide_dir, "tiles")
    spot_directory = os.path.join(slide_dir, "spots")
    tile_metadata_path = os.path.join(slide_dir, "tile_metadata.csv")
    spot_metadata_path = os.path.join(slide_dir, "spot_metadata.csv")
    thumbnail_path = os.path.join(slide_dir, "thumbnail.jpeg")
    annotated_tiles_path = os.path.join(slide_dir, "annotated_thumbnail_tiles.jpeg")
    annotated_spots_path = os.path.join(slide_dir, "annotated_thumbnail_spots.jpeg")
    tissue_mask_path = os.path.join(slide_dir, "tissue_mask.jpeg")
    spot_mask_path = os.path.join(slide_dir, "spot_mask.jpeg")
    unreadable_tiles_path = os.path.join(slide_dir, "unreadable_tiles.csv")
    unreadable_spots_path = os.path.join(slide_dir, "unreadable_spots.csv")
    # Check output directory.
    if os.path.exists(slide_dir):
        if not overwrite:
            return
        for path in [
            tile_metadata_path,
            spot_metadata_path,
            thumbnail_path,
            annotated_spots_path,
            annotated_tiles_path,
            tissue_mask_path,
            spot_mask_path,
            unreadable_tiles_path,
            unreadable_spots_path,
        ]:
            if os.path.exists(path):
                os.remove(path)
        # Then remove tiles and spots.
        if os.path.exists(tile_directory):
            remove_directory(tile_directory)
        if os.path.exists(spot_directory):
            remove_directory(spot_directory)
    # Return all paths.
    return {
        "slide_directory": output_dir,
        "tile_directory": tile_directory,
        "spot_directory": spot_directory,
        "tile_metadata": tile_metadata_path,
        "spot_metadata": spot_metadata_path,
        "thumbnail": thumbnail_path,
        "annotated_tiles": annotated_tiles_path,
        "annotated_spots": annotated_spots_path,
        "tissue_mask": tissue_mask_path,
        "spot_mask": spot_mask_path,
        "unreadable_tiles": unreadable_tiles_path,
        "unreadable_spots": unreadable_spots_path,
    }


def annotate_thumbnail(
    thumbnail: Union[Image.Image, numpy.ndarray],
    thumbnail_downsample: Tuple[float, float],
    coordinates: List[Tuple[int, int, int, int]],
    coordinate_downsample: Tuple[float, float],
    numbers: numpy.ndarray = None,
) -> Image.Image:
    """Annotate thumbnail with tile coordinates."""
    # Draw tiles to the thumbnail.
    if isinstance(thumbnail, numpy.ndarray):
        thumbnail = Image.fromarray(thumbnail)
    # Make a copy.
    thumbnail = thumbnail.copy()
    annotated = ImageDraw.Draw(thumbnail)
    # Get downsample.
    downsample = tuple(
        a / b for a, b in zip(thumbnail_downsample, coordinate_downsample)
    )
    # Start drawing.
    for x, y, w, h in coordinates:
        y_d = round(y / downsample[0])
        h_d = round(h / downsample[0])
        x_d = round(x / downsample[1])
        w_d = round(w / downsample[1])
        # xy expects [(x0, y0), (x1, y1)].
        annotated.rectangle(
            xy=[(x_d, y_d), (x_d + w_d, y_d + h_d)],
            outline="red",
            width=2,
        )
    if numbers is not None:
        arr = numpy.array(thumbnail)
        for (x, y, w, h), number in zip(coordinates, numbers):
            y_d = round(y / downsample[0])
            h_d = round(h / downsample[0])
            x_d = round(x / downsample[1])
            w_d = round(w / downsample[1])
            # Write text.
            cv2.putText(
                img=arr,
                text=number,
                org=(x_d, y_d + h_d),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                fontScale=1,
                color=(0, 0, 255),
                thickness=2,
            )
        thumbnail = Image.fromarray(arr)
    return thumbnail


def worker_initializer_fn(slide_reader):
    """Save slide reader as a global variable."""
    global __SLIDE_READER__
    __SLIDE_READER__ = slide_reader


def save_tile_worker(
    xywh: Tuple[int, int, int, int],
    level: int,
    tile_directory: str,
    tissue_threshold: int,
    preprocess_metrics: Callable,
    image_format: str,
    quality: int,
) -> Tuple[tuple, dict, float]:
    """Save a tile image and calculate preprocessing metrics. If a region cannot
    be read, returns the Exception instead.

    NOTE: Requires that each worker is initialized with `worker_initializer_fn`.
    """
    tic = time.perf_counter()
    # Unpack.
    x, y, w, h = xywh
    # Save basic metadata.
    tile_metadata = {
        "slide_name": __SLIDE_READER__.slide_name,
        "x": x,
        "y": y,
        "w": w,
        "h": h,
        "level": level,
    }
    # Read region.
    try:
        tile = __SLIDE_READER__.read_region((x, y, w, h), level=level)
    except Exception as e:
        # Could not read region -> return the Exception and times as None.
        return xywh, e, None
    # Define tile path.
    filepath = os.path.join(
        tile_directory, "x{}_y{}_w{}_h{}.{}".format(x, y, w, h, image_format)
    )
    tile_metadata["path"] = os.path.realpath(filepath)
    # Update metadata with preprocessing metrics.
    if preprocess_metrics is not None:
        try:
            tile_metadata.update(preprocess_metrics(tile, tissue_threshold))
        except Exception as e:
            # Could not calculate preprocessing metrics.
            return xywh, e, None
    # Save image.
    if not os.path.exists(filepath):
        try:
            tile.save(filepath, quality=quality)
        except Exception as e:
            # Could not save region.
            return xywh, e, None
    # Return metadata and times.
    return xywh, tile_metadata, time.perf_counter() - tic


def save_spot_worker(
    spot_info: Tuple[int, int, int, int, str],
    spot_directory: str,
    image_format: str,
    quality: int,
):
    """Saves a single spot."""
    tic = time.perf_counter()
    # Unpack.
    x, y, w, h, number = spot_info
    # Create name.
    spot_name = "{}_spot_{}".format(__SLIDE_READER__.slide_name, number)
    # Save basic metadata.
    spot_metadata = {
        "slide_name": __SLIDE_READER__.slide_name,
        "x": x,
        "y": y,
        "w": w,
        "h": h,
        "spot_number": number,
        "spot_name": spot_name,
    }
    # Read region.
    try:
        spot = __SLIDE_READER__.read_region((x, y, w, h))
    except Exception as e:
        # Could not read region.
        return spot_info, e, None
    # Define spot path.
    filepath = os.path.join(spot_directory, "{}.{}".format(spot_name, image_format))
    spot_metadata["path"] = filepath
    # Save image.
    if not os.path.exists(filepath):
        try:
            spot.save(filepath, quality=quality)
        except Exception as e:
            # Could not save spot.
            return spot_info, e, None
    # Return metadata and times.
    return spot_info, spot_metadata, time.perf_counter() - tic



from typing import Dict, Tuple

import aicspylibczi
import cv2
import numpy

from ._base import Backend

__all__ = ["ZeissBackend"]

ZEISS_READABLE = ("czi",)


class ZeissBackend(Backend):
    def __init__(self, path: str):
        self.__path = path
        self.reader = aicspylibczi.CziFile(path)
        if not self.reader.is_mosaic():
            raise NotImplementedError(
                "HistoPrep does not support non-mosaic zeiss files."
            )
        # Get plane constraints.
        bbox = self.reader.get_mosaic_bounding_box()
        self.__dimensions = (bbox.h, bbox.w)
        self.__origo = (bbox.x, bbox.y)
        # Init caches.
        self.__level_dimensions = None
        self.__level_downsamples = None

    def get_dimensions(self) -> Tuple[int, int]:
        return self.__dimensions

    def get_level_dimensions(self) -> Dict[int, Tuple[int, int]]:
        if self.__level_dimensions is None:
            level_dimensions = {0: self.__dimensions}
            downsample = 1
            while True:
                if max(self.dimensions) // 2**downsample < 512:
                    break
                level_dimensions[downsample] = tuple(
                    x // 2**downsample for x in self.dimensions
                )
                downsample += 1
            self.__level_dimensions = level_dimensions
        return self.__level_dimensions

    def get_level_downsamples(self) -> Dict[int, Tuple[int, int]]:
        if self.__level_downsamples is None:
            level_downsamples = {}
            for level, (y, x) in self.level_dimensions.items():
                level_downsamples[level] = (
                    self.dimensions[0] / y,
                    self.dimensions[1] / x,
                )
            self.__level_downsamples = level_downsamples
        return self.__level_downsamples

    def get_thumbnail(self, level: int) -> numpy.ndarray:
        # Calculate scale factor.
        scale_factor = 1 / 2**level
        # Read thumbnail.
        thumbnail = self.reader.read_mosaic(
            scale_factor=scale_factor,
            C=0,
            background_color=(1.0, 1.0, 1.0),
        )[0]
        # Convert BGR to RGB.
        return cv2.cvtColor(thumbnail, cv2.COLOR_BGR2RGB)

    def read_region(self, XYWH: Tuple[int, int, int, int], level: int) -> numpy.ndarray:
        # Unpack.
        x, y, w, h = XYWH
        # Calculate scale factor.
        scale_factor = 1 / 2**level
        # Fix coordinates.
        XYWH = (
            self.__origo[0] + int(x / scale_factor),
            self.__origo[1] + int(y / scale_factor),
            int(w / scale_factor),
            int(h / scale_factor),
        )
        # Read region.
        tile = self.reader.read_mosaic(
            region=XYWH,
            scale_factor=scale_factor,
            C=0,
            background_color=(1.0, 1.0, 1.0),
        )[0]
        # Convert BGR to RGB.
        return cv2.cvtColor(tile, cv2.COLOR_BGR2RGB)

    def __repr__(self):
        return "ZEISS"



import setuptools

exec(open("histoprep/_version.py").read())

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="histoprep",
    version=__version__,
    author="jopo666",
    author_email="jopo@birdlover.com",
    description="Preprocessing module for large histological images.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jopo666/HistoPrep",
    scripts=["HistoPrep"],
    packages=setuptools.find_packages(include=["histoprep", "histoprep.*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords="image-analysis preprocessing histology openslide pathology",
    python_requires=">=3.8",
    install_requires=[
        "aicspylibczi==3.0.5",
        "numpy==1.22",
        "pandas==1.4",
        "Pillow==9.0",
        "openslide-python==1.1.2",
        "opencv-python==4.5",
        "scikit-learn==1.1",
        "matplotlib==3.5",
    ],
)



import os
import shutil

import pandas
from histoprep.helpers import combine_metadata, strip_metric_colums

from paths import DATA_PATH, TMP_DIR, clean_tmp_dir


def test_strip_metrics():
    meta = pandas.read_csv(os.path.join(DATA_PATH, "metadata.csv"))
    assert meta.columns.tolist() == [
        "slide_name",
        "x",
        "y",
        "w",
        "h",
        "path",
        "background",
        "gray_std",
        "red_std",
        "green_std",
        "blue_std",
        "hue_std",
        "saturation_std",
        "brightness_std",
        "black_pixels",
        "white_pixels",
        "sharpness_max",
        "red_q=0.05",
        "red_q=0.1",
        "red_q=0.5",
        "red_q=0.9",
        "red_q=0.95",
        "green_q=0.05",
        "green_q=0.1",
        "green_q=0.5",
        "green_q=0.9",
        "green_q=0.95",
        "blue_q=0.05",
        "blue_q=0.1",
        "blue_q=0.5",
        "blue_q=0.9",
        "blue_q=0.95",
        "hue_q=0.05",
        "hue_q=0.1",
        "hue_q=0.5",
        "hue_q=0.9",
        "hue_q=0.95",
        "saturation_q=0.05",
        "saturation_q=0.1",
        "saturation_q=0.5",
        "saturation_q=0.9",
        "saturation_q=0.95",
        "brightness_q=0.05",
        "brightness_q=0.1",
        "brightness_q=0.5",
        "brightness_q=0.9",
        "brightness_q=0.95",
        "scanner",
    ]
    assert strip_metric_colums(meta).columns.tolist() == [
        "slide_name",
        "x",
        "y",
        "w",
        "h",
        "path",
        "background",
        "scanner",
    ]


def test_combine_metadata():
    metadata_path = os.path.join(DATA_PATH, "metadata.csv")
    clean_tmp_dir()
    for i in range(5):
        slide_dir = os.path.join(TMP_DIR, "slide_{}".format(i))
        os.makedirs(slide_dir)
        slide_metadata_path = os.path.join(slide_dir, "metadata.csv")
        shutil.copy(metadata_path, slide_metadata_path)
    original = pandas.read_csv(metadata_path)
    combined = combine_metadata(TMP_DIR, "metadata.csv")
    assert len(original) * 5 == len(combined)
    clean_tmp_dir()



from typing import List, Tuple, Union

import numpy
from PIL import Image

from ._functional import _rgb2hsv, _to_grayscale

ALLOWED_REDUCTIONS = ["max", "median", "mean", "min"]

__all__ = [
    "downsample_image",
    "resize_image",
    "arr2pil",
    "rgb2hsv",
    "rgb2gray",
]


def _to_array(image: Union[Image.Image, numpy.ndarray]) -> numpy.ndarray:
    """Convert image to array and check it's a valid image."""
    if isinstance(image, Image.Image):
        image = numpy.array(image)
    elif not isinstance(image, numpy.ndarray):
        raise TypeError(
            "Input image should be an array or image not {}".format(type(image))
        )
    # Check image.
    if image.dtype != numpy.uint8:
        raise TypeError("Image dtype should be uint8 not {}.".format(image.dtype))
    if not image.ndim >= 2:
        raise TypeError("Number of dimensions is too low for an image")
    if image.ndim != 2 and image.shape[-1] != 3:
        raise TypeError(
            "Image should have 1 or 3 channels but found: {}.".format(image.shape[-3])
        )
    return image


def _five_crop(image: numpy.ndarray) -> List[numpy.ndarray]:
    """Take five crops from the image."""
    height, width = image.shape[:2]
    crop_height, crop_width = height // 2, width // 2
    # Corners.
    tl = image[:crop_height, :crop_width]
    tr = image[:crop_height, crop_width:]
    bl = image[crop_height:, :crop_width]
    br = image[crop_height:, crop_width:]
    # Center.
    center_h = crop_height // 2
    center_w = crop_width // 2
    center = image[center_h : crop_height + center_h, center_w : crop_width + center_w]
    return tl, tr, bl, br, center


def _reduce_values(values: numpy.ndarray, method: str) -> float:
    """Reduce values of values with given method."""
    if method not in ALLOWED_REDUCTIONS:
        raise ValueError(
            "Reduction %s not recognised. Select from %s."
            % (method, ALLOWED_REDUCTIONS)
        )
    if method == "max":
        reduced = numpy.max(values)
    elif method == "median":
        reduced = numpy.median(values)
    elif method == "mean":
        reduced = numpy.mean(values)
    elif method == "min":
        reduced = numpy.min(values)
    return float(reduced)


def rgb2gray(
    image: Union[numpy.ndarray, Image.Image]
) -> Union[numpy.ndarray, Image.Image]:
    """Convert an RGB image to grayscale.

    Args:
        image: RGB image.

    Returns:
        Grayscale image.
    """
    gray = _to_grayscale(_to_array(image))
    if isinstance(image, Image.Image):
        gray = Image.fromarray(gray)
    return gray


def rgb2hsv(
    image: Union[numpy.ndarray, Image.Image]
) -> Union[numpy.ndarray, Image.Image]:
    """Convert an RGB image to HSV.

    Args:
        image: RGB image.

    Returns:
        HSV image.
    """
    if image.ndim == 2:
        raise ValueError("Image has too few dimensions to bee an RGB image.")
    hsv = _rgb2hsv(_to_array(image))
    if isinstance(image, Image.Image):
        hsv = Image.fromarray(hsv)
    return hsv


def arr2pil(
    image: Union[numpy.ndarray, Image.Image],
    equalize: bool = False,
) -> Image.Image:
    """Convert numpy array to a Pillow image.

    Args:
        image: Image array.
        equalize: Equalise image, useful for plotting masks.

    Returns:
        Pillow image.

    Example:
        ```python
        import histoprep.functional as F
        from histoprep.helpers import read_image

        image = read_image("path/to/image.jpeg")
        tissue_mask = F.detect_tissue(image)
        # Create nice mask image for visualization.
        tissue_mask_pil = F.arr2pil(1 - tissue_mask, equalize=True)
        ```
    """
    image = _to_array(image)
    if equalize and image.max() != 0:
        image = (image / image.max()) * 255
    image = Image.fromarray(image.astype(numpy.uint8))
    return image


def downsample_image(
    image: Union[numpy.ndarray, Image.Image],
    max_dimension: int = 2048,
    return_arr: bool = False,
) -> Union[Image.Image, numpy.ndarray]:
    """
    Donwsaple image until one of the dimensions is less than ``max_dimension``.

    Args:
        image: Image to be downsampled.
        max_dimension: Maximum dimension size. Defaults to 2048.

    Returns:
        Downsampled image.

    Example:
        ```python
        import histoprep.functional as F
        from histoprep.helpers import read_image

        image = read_image("path/to/image.jpeg")
        downsampled = F.downsample_image(image, max_dimension=32)
        assert all(x<=32 for x in downsampled.size)
        ```
    """
    image = _to_array(image)
    height, width = image.shape[:2]
    factor = 0
    while width > max_dimension or height > max_dimension:
        factor += 1
        height = int(height / 2**factor)
        width = int(width / 2**factor)
    image = resize_image(image, (height, width), fast_resize=True)
    if not return_arr:
        return image
    else:
        return numpy.array(image)


def resize_image(
    image: Union[Image.Image, numpy.ndarray],
    shape: Union[int, Tuple[int]],
    return_arr: bool = False,
    fast_resize: bool = False,
) -> Union[Image.Image, numpy.ndarray]:
    """Resize image to desired shape.

    Args:
        image: Input image.
        shape: Shape of the output image. If shape is an integer then the image
            is resized to a square of (shape, shape).
        return_arr: Return array. Defaults to False.
        fast_resize: Uses nearest neigbour interpolation. Defaults to False.

    Returns:
        Resized image.

    Example:
        ```python
        import histoprep.functional as F
        from histoprep.helpers import read_image

        image = read_image("path/to/image.jpeg")
        resized_32 = F.resize_image(image, (32, 32))
        resized_32 = F.resize_image(image, 32)
        resized_32_arr = F.resize_image(image, 32, return_arr=True)
        ```
    """
    if isinstance(image, numpy.ndarray):
        image = Image.fromarray(image)
    if isinstance(shape, int):
        shape = (shape, shape)
    if fast_resize:
        image = image.resize(shape[::-1], 0)
    else:
        image = image.resize(
            shape[::-1],
        )
    if not return_arr:
        return image
    else:
        return numpy.array(image)



import functools
import logging
import os
import warnings
from typing import Callable, Dict, Iterable, List, Tuple, Union

import numpy
import pandas
from PIL import Image

from .. import functional as F
from ..helpers._files import get_extension, remove_extension
from ..helpers._multiprocess import multiprocess_loop
from ..helpers._verbose import progress_bar, verbose_fn
from ._backend import (
    OPENSLIDE_READABLE,
    PILLOW_READABLE,
    READABLE_FORMATS,
    ZEISS_READABLE,
    OpenSlideBackend,
    PillowBackend,
    ZeissBackend,
)
from ._utils import (
    annotate_thumbnail,
    check_level,
    check_region,
    find_level,
    prepare_paths_and_directories,
    save_spot_worker,
    save_tile_worker,
    worker_initializer_fn,
)

__all__ = ["SlideReader"]

ALLOWED_IMAGE_FORMATS = ("png", "jpeg", "jpg")


class SlideReader(object):
    def __init__(
        self,
        path: str,
        verbose: bool = False,
        preferred_dimension: int = 4096,
        threshold: int = None,
        threshold_multiplier: float = 1.05,
        max_dimension: int = 2**14,
    ):
        """Reader for large slide images with extended functionality such as
        tissue detection, tile coordinate extraction, dearraying TMA spots and
        saving tiles.

        Args:
            path: Path to slide image.
            verbose: Set to True to get verbose output. Defaults to False.
            preferred_dimension: Preferred maximum dimension for the for the
                generated thumbnail. Defaults to 4096.
            threshold: Treshold for tissue detection. If set, will detect tissue
                by global thresholding, and otherwise Otsu's method is used to
                find a threshold. Defaults to None.
            threshold_multiplier: Otsu's method is used to find an optimal
                threshold by minimizing the weighted within-class variance. This
                threshold is then multiplied with `threshold_multiplier`. Used
                only if `threshold` is None. Defaults to 1.05.
            max_dimension: Maximum dimension for the generated thumbnail.
                Smaller thumbnails are preferred, but in the absence of these,
                the first level with all dimensions less than `max_dimension`
                is chosen. Defaults to 16_384.

        Raises:
            FileNotFoundError: File not found.
            NotADirectoryError: Scratch directory exists, but isn't a directory.
            IOError: File not readable by HistoPrep.

        Usage:
        ```python
        from histoprep import SlideReader

        # Load tile images.
        reader = SlideReader("/path/to/slide.tiff")
        thumbnail = reader.get_thumbnail()
        thresh, tissue_mask = reader.detect_tissue()
        coords = reader.get_tile_coordinates(
            width=512, overlap=0.1, tissue_mask=tissue_mask
        )
        tiles = [reader.read_region(xywh) for xywh in coords]
        ```
        """
        # Check path.
        if not os.path.exists(path):
            raise FileNotFoundError("{} does not exist.".format(path))
        elif not os.path.isfile(path):
            raise IOError("{} is not a file.".format(path))
        # Load slide with the correct backend.
        extension = get_extension(path)
        if extension is None:
            raise IOError("{} file has no extension.".format(path))
        elif extension in OPENSLIDE_READABLE:
            logging.debug("Using OPENSLIDE backend.")
            self.__backend = OpenSlideBackend(path)
        elif extension in PILLOW_READABLE:
            logging.debug("Using PILLOW backend.")
            self.__backend = PillowBackend(path)
        elif extension in ZEISS_READABLE:
            logging.debug("Using ZEISS backend.")
            self.__backend = ZeissBackend(path)
            self.__max_workers = max(1, os.cpu_count() // 2)
        else:
            raise ValueError(
                "Extension {} is not readable with HistoPrep. Supported file "
                "formats are: {}".format(extension, READABLE_FORMATS)
            )
        # Save attributes.
        self.__path = path
        self.__slide_name = remove_extension(os.path.basename(path))
        self.__max_workers = os.cpu_count()
        self.__max_dimension = max_dimension
        # Define verbose function.
        self.__verbose = functools.partial(
            verbose_fn, verbose=verbose, desc=self.__slide_name, color=True
        )
        # Generate thumbnail and detect tissue.
        self.get_thumbnail(preferred_dimension)
        self.detect_tissue(threshold=threshold, multiplier=threshold_multiplier)
        self.__annotated_thumbnail_tiles = None
        self.__annotated_thumbnail_spots = None
        self.__tile_metadata = None
        self.__spot_metadata = None
        self.__spot_mask = None
        self.__spot_info = None

    @property
    def MAX_DIMENSION(self) -> int:
        return self.__max_dimension

    # a setter function
    @MAX_DIMENSION.setter
    def MAX_DIMENSION(self, value: int):
        if not isinstance(value, int):
            raise TypeError("Maximum dimension should be an integer.")
        if value < 0:
            raise ValueError("Maximum dimension should be positive.")
        self.__max_dimension = value

    @property
    def path(self) -> str:
        return self.__path

    @property
    def slide_name(self) -> str:
        return self.__slide_name

    @property
    def backend(self) -> str:
        return self.__backend

    @property
    def channel_order(self) -> str:
        return "XYWH"

    @property
    def dimension_order(self) -> str:
        return "HW"

    @property
    def dimensions(self) -> Tuple[int, int]:
        """Slide image dimensions."""
        return self.__backend.dimensions

    @property
    def level_downsamples(self) -> Dict[int, Tuple[float, float]]:
        """Dimensions for each level."""
        return self.__backend.level_downsamples

    @property
    def level_dimensions(self) -> Dict[int, Tuple[int, int]]:
        """Downsamples for each level."""
        return self.__backend.level_dimensions

    @property
    def tissue_threshold(self) -> int:
        """Threshold for tissue detection."""
        return self.__threshold

    @property
    def thumbnail(self) -> Image.Image:
        """Thumbnail image of the slide."""
        return self.__thumbnail

    @property
    def thumbnail_downsample(self) -> Image.Image:
        """Downsample for the thumbnail image."""
        return self.level_downsamples[self.__thumbnail_level]

    @property
    def tissue_mask(self) -> Image.Image:
        """Tissue mask (0=background, 1=tissue)."""
        return F.arr2pil(1 - self.__tissue_mask, equalize=True)

    @property
    def spot_mask(self) -> Image.Image:
        """TMA spot mask (0=background, 1=spot)."""
        if self.__spot_mask is not None:
            return F.arr2pil(1 - self.__spot_mask, equalize=True)

    @property
    def annotated_thumbnail_tiles(self) -> Image.Image:
        """Thumbnail image annotated with the tile coordinates."""
        return self.__annotated_thumbnail_tiles

    @property
    def annotated_thumbnail_spots(self) -> Image.Image:
        """Thumbnail image annotated with the spot coordinates."""
        return self.__annotated_thumbnail_spots

    @property
    def tile_metadata(self) -> Image.Image:
        """Metadata for saved tiles."""
        return self.__tile_metadata

    @property
    def spot_metadata(self) -> Image.Image:
        """Metadata for TMA spots."""
        return self.__spot_metadata

    def read_region(
        self,
        xywh: Tuple[int, int, int, int],
        level: int = 0,
        return_arr: bool = False,
        fill: int = 255,
    ) -> Union[Image.Image, numpy.ndarray]:
        """Read region from slide.

        Args:
            xywh: Region defined by X, Y, width and height.
            return_arr: Return region as an uint8 array instead of a Pil image.
                Defaults to False.
            fill: Fill overbound areas with this value. If None, raises an error
                when the region goes out of bounds. Defaults to 255.

        Raises:
            ValueError: XYWH, level or fill are not integers.
            ValueError: Negative coordinates.
            ValueError: Width or height less than 1.
            ValueError: Region overbound without a fill value.

        Returns:
            Image of the region.

        Example:
            ```python
            from histoprep import SlideReader

            slide = SlideReader("/path/to/slide.tiff")
            tile_image = slide.read_region(x=0, y=0, w=1024, h=1024)
            ```
        """
        if any(not isinstance(z, int) for z in xywh if z):
            raise TypeError("XYWH coordinates should be integers.")
        x, y, width, height = xywh
        if x < 0 or y < 0:
            raise ValueError(
                "Coordinates should be positive, not ({},{}).".format(x, y)
            )
        if width < 1 or height < 1:
            raise ValueError("Width and height should be above zero.")
        if fill and (not isinstance(fill, int) or not 0 <= fill <= 255):
            raise ValueError("Fill value should be an integer in range [0, 255].")
        out_of_bounds, padding = check_region(xywh, self.dimensions)
        if level not in self.level_downsamples.keys():
            raise ValueError("Slide does not contain level {}.".format(level))
        if out_of_bounds:
            # Tile goes out-of-bounds.
            if fill is None:
                raise ValueError(
                    "Region ({}) overbound and fill is None (dimensions: "
                    "{}).".format((x, y, width, height), self.dimensions)
                )
            if padding is None:
                # Empty tile.
                tile = numpy.zeros((height, width), dtype=numpy.uint8) + fill
            else:
                # Read allowed region.
                allowed_h, allowed_w = padding
                tile = self.__backend.read_region(
                    XYWH=(x, y, allowed_w, allowed_h),
                    level=level,
                )
                # Pad tile.
                tile = numpy.pad(
                    tile,
                    ((0, height - allowed_h), (0, width - allowed_w), (0, 0)),
                    constant_values=fill,
                )
        else:
            # Read region.
            tile = self.__backend.read_region(XYWH=(x, y, width, height), level=level)
        if not return_arr:
            tile = Image.fromarray(tile)
        # Return tile
        return tile

    def get_thumbnail(
        self,
        preferred_dimension: int = 4096,
        level: int = None,
        return_arr: bool = False,
        fill: int = 255,
        return_level: bool = False,
    ) -> Union[numpy.ndarray, Image.Image]:
        """Get thumbnail image of the slide.

        Running this method updates the `thumbnail` attribute.

        Args:
            preferred_dimension: Selects the first level which has both
                dimensions smaller than `preferred_dimension`. If not found,
                selects the smallest thumbnail with all dimensions less than
                `MAX_DIMENSION` attribute. Defaults to 4096.
            level: Overwrites max_dimension and loads a specific level. Defaults
                to None.
            return_arr: Return an uint8 array instead of PIL Image. Defaults to
                False.
            fill: Fill out of bounds regions. Defaults to 255.
            return_level: Return the thumbnail level. Defaults to False.

        Raises:
            ValueError: No levels with dimensions less than `MAX_DIMENSION`.
            ValueError: Level does not exist.

        Returns:
            Thumbnail image.

        Example:
            ```python
            from histoprep import SlideReader

            reader = Slidereader("/path/to/slide.tiff")
            thumbnail = reader.get_thumbnail()
            ```
        """
        if level is None and preferred_dimension is None:
            raise ValueError("Either level or max_dimension must be defined.")
        if level is None:
            level = find_level(
                preferred_dimension,
                self.MAX_DIMENSION,
                self.level_dimensions,
            )
        else:
            check_level(level, self.level_dimensions)
        # Load thumbnail.
        thumbnail = self.__backend.get_thumbnail(level)
        if fill is not None:
            # Fill completely black pixels.
            thumbnail[thumbnail.sum(-1) == 0] = fill
        if not return_arr:
            # Convert to PIL.
            thumbnail = Image.fromarray(thumbnail)
        downsample = self.level_downsamples[level]
        if isinstance(downsample, (tuple, list)):
            downsample = "({})".format(
                ", ".join(["{:.3f}".format(x) for x in downsample])
            )
        else:
            downsample = "{:.3f}".format(downsample)
        self.__verbose(
            "Thumbnail created with downsample {}.".format(downsample),
        )
        # Cache.
        self.__thumbnail = thumbnail
        self.__thumbnail_level = level
        if return_level:
            return level, thumbnail
        else:
            return thumbnail

    def detect_tissue(
        self,
        threshold: int = None,
        multiplier: float = 1.05,
        sigma: float = 1.0,
        remove_white: bool = True,
    ):
        """Detect tissue from slide image.

        Running this method updates the `tissue_mask` attribute.

        Args:
            threshold: Treshold for tissue detection. If set, will detect tissue
                by global thresholding, and otherwise Otsu's method is used to
                find a threshold. Defaults to None.
            multiplier: Otsu's method is used to find an optimal threshold by
                minimizing the weighted within-class variance. This threshold is
                then multiplied with `multiplier`. Used only if `threshold` is
                None. Defaults to 1.0.
            sigma: Sigma for gaussian blurring. Defaults to 1.05.
            remove_white: Changes completely white regions to the second highest
                pixel value, before detecting background. This way Otsu's method
                isn't affected by possibly extreme white areas. Defaults to
                True.

        Returns:
            Binary mask with 0=background and 1=tissue.

        Example:
            ```python
            from histoprep import SlideReader

            reader = SlideReader("path/to/slide")
            thresh, tissue_mask = reader.detect_tissue()
            ```
        """
        # Detect tissue.
        threshold, tissue_mask = F.detect_tissue(
            image=self.__thumbnail,
            threshold=threshold,
            multiplier=multiplier,
            sigma=sigma,
            remove_white=remove_white,
        )
        self.__verbose(
            "Tissue detected with threshold {}.".format(threshold),
        )
        # Cache.
        self.__tissue_mask = tissue_mask
        self.__threshold = threshold
        return threshold, tissue_mask

    def get_tile_coordinates(
        self,
        width: int,
        overlap: float = 0.0,
        height: int = None,
        level: int = 0,
        tissue_mask: numpy.ndarray = None,
        max_background: float = 0.95,
    ) -> List[Tuple[int, int, int, int]]:
        """Get tile coordinates for the slide.

        Divides the slide into tiles based on width, height and overlap. These
        tiles are then filtered based on the amount of background. Running this
        method updates the `annotated_thumbnail_tiles` attribute.

        Args:
            width: Tile width.
            overlap: Overlap between neighbouring tiles. Defaults to 0.0.
            level: Slide level for calculating tile coordinates. Defaults to 0.
            height: Height of a tile. If None, will be set to width. Defaults to
                None.
            tissue_mask: Tissue mask. If None, uses a cached tissue mask.
                Defaults to None.
            max_background: Maximum amount of background in tile. Defaults to
                0.95.

        Returns:
            XYWH coordinates for each tile.
        """
        if level not in self.level_dimensions.keys():
            raise ValueError("Slide does not contain level {}.".format(level))
        else:
            dimensions = self.level_dimensions[level]
        coordinates = F.tile_coordinates(
            dimensions, width=width, height=height, overlap=overlap
        )
        # Filter tiles.
        if tissue_mask is None:
            # Use cached.
            tissue_mask = self.__tissue_mask
        # Figure out downsample.
        downsample = [b / a for a, b in zip(tissue_mask.shape[:2], dimensions)]
        # Filter coordinates.
        filtered = F.filter_coordinates(
            coordinates=coordinates,
            tissue_mask=tissue_mask,
            max_background=max_background,
            downsample=downsample,
        )
        self.__verbose(
            "Detected {} tiles with <= {:.1f}{} background "
            "(filtered {:.1f}{} of all tiles).".format(
                len(filtered),
                100 * max_background,
                "%",
                100 - 100 * len(filtered) / len(coordinates),
                "%",
            )
        )
        # Annotate thumbnail and cache it.
        self.__annotated_thumbnail_tiles = annotate_thumbnail(
            thumbnail=self.thumbnail,
            thumbnail_downsample=self.thumbnail_downsample,
            coordinates=filtered,
            coordinate_downsample=self.level_downsamples[level],
        )
        return filtered

    def dearray(
        self,
        kernel_size: int = 5,
        iterations: int = 3,
        min_area: float = 0.25,
        max_area: float = 3.0,
    ) -> pandas.DataFrame:
        """Dearray TMA spots.

        Running this method updates the `annotated_thumbnail_spots` and
        `spot_mask` attributes.

        Args:
            kernel_size: Kernel size for dilate. Defaults to 5.
            iterations: Dilate iterations. Defaults to 3.
            min_area: Minimum area for a spot. Defaults to 0.1 and calculated
                with: `median(areas) * min_area`
            max_area: Maximum area for a spot. Similar to min_area. Defaults to
                3.0.

        Returns:
            Spot metadata.
        """
        # Dearray.
        self.__spot_mask, bboxes, spot_numbers = F.dearray(
            tissue_mask=self.__tissue_mask,
            kernel_size=kernel_size,
            iterations=iterations,
            min_area=min_area,
            max_area=max_area,
        )
        # Figure out downsample.
        h_d, w_d = [
            b / a for a, b in zip(self.__tissue_mask.shape[:2], self.dimensions)
        ]
        # Multiply bboxes to get actual coordinates.
        spot_coordinates = []
        for (x, y, w, h) in bboxes:
            # Multiply with downsample.
            spot_coordinates.append(
                (round(x * w_d), round(y * h_d), round(w * w_d), round(h * h_d))
            )
        # Annotate thumbnail.
        self.__annotated_thumbnail_spots = annotate_thumbnail(
            thumbnail=self.thumbnail,
            thumbnail_downsample=self.thumbnail_downsample,
            coordinates=spot_coordinates,
            coordinate_downsample=(1, 1),
            numbers=spot_numbers,
        )
        # Create metadata.
        rows = []
        for xywh, number in zip(spot_coordinates, spot_numbers):
            rows.append((self.slide_name, *xywh, number))
        self.__spot_metadata = pandas.DataFrame(rows)
        self.__spot_metadata.columns = [
            "slide_name",
            "x",
            "y",
            "w",
            "h",
            "spot_number",
        ]
        # Create simple info for saving.
        self.__spot_info = []
        for __, row in self.spot_metadata.iterrows():
            self.__spot_info.append(row[["x", "y", "w", "h", "spot_number"]].tolist())
        return self.__spot_metadata

    def __prepare_for_saving_images(
        self,
        output_dir: str,
        overwrite: bool = False,
        image_format: str = "jpeg",
        quality: int = 95,
        num_workers: int = None,
    ):
        # Check image formats.
        if image_format.lower() not in ALLOWED_IMAGE_FORMATS:
            raise ValueError(
                "Image format %s not allowed. Select from %s."
                % (image_format, ALLOWED_IMAGE_FORMATS)
            )
        # Check quality.
        if image_format.lower() in {"jpeg", "jpg"} and not 0 < quality <= 100:
            raise ValueError("Image quality should be in range (0,100].")
        # Define number of workers.
        num_workers = os.cpu_count() if num_workers is None else num_workers
        num_workers = max(min(self.__max_workers, num_workers), 1)
        # Define outputs paths and directories.
        output_paths = prepare_paths_and_directories(
            output_dir=output_dir,
            slide_name=self.slide_name,
            overwrite=overwrite,
        )
        return num_workers, output_paths

    def __save_images(
        self,
        worker_function: Callable,
        iterable: Iterable,
        num_workers: int,
        num_retries: int,
        csv_path: str,
        unreadable_path: str,
        display_progress: bool,
    ) -> pandas.DataFrame:
        # Initialize lists for metadata and unreadable regions.
        metadata = []
        failed_regions = []
        # Initialise the main worker.
        worker_initializer_fn(self)
        # Define loop.
        if num_workers < 2:
            # Single process.
            loop = map(worker_function, iterable)
        else:
            # Multiple processes.
            loop = multiprocess_loop(
                func=worker_function,
                iterable=iterable,
                num_workers=num_workers,
                initializer=worker_initializer_fn,
                initializer_args=(self,),
            )
        # Init metadata and failed regions.
        metadata = []
        failed_regions = []
        unreadable = []
        # Save images.
        self.__verbose("Saving images with {} workers.".format(num_workers))
        for logs, (info, meta, elapsed) in progress_bar(
            iterable=loop,
            total=len(iterable),
            log_values=True,
            suppress=not display_progress,
        ):
            if isinstance(meta, Exception):
                # Region couldn't be read, preprocessed or saved.
                logging.debug("Region ({}) raised exception: {}".format(info[:4], meta))
                failed_regions.append((info[:4], meta))
                logs["failed"] = len(failed_regions)
                continue
            # Append to metadata.
            metadata.append(meta)
            # Log elapsed time.
            if elapsed < 1:
                logs["per_image"] = "{:.1f}ms".format(1000 * elapsed)
            else:
                logs["per_image"] = "{:.1f}s".format(elapsed)
        # Attempt to read unreadable regions again.
        if len(failed_regions) > 0 and num_retries > 0:
            self.__verbose(
                "Attempting to read {} failed regions again.".format(
                    len(failed_regions)
                )
            )
            num_workers = num_workers // 2
            if num_workers < 2:
                # Single process.
                retry_loop = map(
                    worker_function,
                    (info for info, __ in failed_regions),
                )
            else:
                # Multiple processes.
                retry_loop = multiprocess_loop(
                    func=worker_function,
                    iterable=(info for info, __ in failed_regions),
                    num_workers=num_workers,
                    initializer=worker_initializer_fn,
                    initializer_args=(self,),
                )
            # Try to save images again.
            for info, meta, __ in progress_bar(
                iterable=retry_loop,
                total=len(failed_regions),
                suppress=not display_progress,
                desc="Retrying failed regions",
            ):
                if isinstance(meta, Exception):
                    for i in range(num_retries):
                        __, meta, __ = worker_function(info)
                        if isinstance(meta, dict):
                            # Success!
                            break
                if isinstance(meta, Exception):
                    unreadable.append((info[:4], meta))
                    logs["unreadable"] = len(unreadable)
                    continue
                # Append to metadata if reading succeeded.
                metadata.append(meta)
            if len(unreadable) > 0:
                self.__verbose(
                    "Couldn't read {} regions. Saving information to {}".format(
                        len(unreadable), unreadable_path
                    )
                )
                unreadable = pandas.DataFrame(unreadable)
                unreadable.columns = ["xywh", "exception"]
                # Save.
                unreadable.to_csv(unreadable_path, index=False)
        # Combine metadata.
        if len(metadata) > 0:
            metadata = pandas.DataFrame(metadata)
            metadata = metadata.sort_values(by=["x", "y"])
            # Save metadata.
            self.__verbose("Saving metadata to {}".format(csv_path))
            metadata.to_csv(csv_path, index=False)
            return metadata
        else:
            warnings.warn("No tiles saved from {}.".format(self.slide_name))

    def save_spots(
        self,
        output_dir: str,
        overwrite: bool = False,
        image_format: str = "jpeg",
        quality: int = 95,
        num_workers: int = os.cpu_count(),
        num_retries: int = 10,
        display_progress: bool = True,
    ) -> pandas.DataFrame:
        """Save TMA spot images from a slide.

        Args:
            output_dir: Output directory for the processed images.
            overwrite: Overwrite any existing images. Defaults to False.
            image_format: Image format. Defaults to "jpeg".
            quality: Quality of JPEG compression. Defaults to 95.
            num_workers: Number of image saving workers. Defaults to the number
                of CPU cores.
            num_retries: Number retries loading any regions, which couldn't be
                saved the first time. Defaults to 10.
            display_progress: Display progress bar. Defaults to True.

        Returns:
            Dataframe with spot bounding boxes.
        """
        # Prepare paths and number of workers.
        num_workers, output_paths = self.__prepare_for_saving_images(
            output_dir=output_dir,
            overwrite=overwrite,
            image_format=image_format,
            quality=quality,
            num_workers=num_workers,
        )
        if output_paths is None:
            warnings.warn(
                "Output directory exists but overwrite=False. " "Not saving images."
            )
            return
        # Load spot metadata.
        if self.__spot_info is None:
            self.__verbose("Dearraying slide with default settings.")
            self.dearray()
        # Create output directory.
        os.makedirs(output_paths["spot_directory"], exist_ok=True)
        self.__verbose("Saving annotated spot thumbnail and spot mask.")
        # Save thumbnail and mask.
        self.thumbnail.save(output_paths["thumbnail"])
        self.annotated_thumbnail_spots.save(output_paths["annotated_spots"])
        self.spot_mask.save(output_paths["spot_mask"])
        # Save spots.
        self.__spot_metadata = self.__save_images(
            worker_function=functools.partial(
                save_spot_worker,
                **{
                    "spot_directory": output_paths["spot_directory"],
                    "image_format": image_format,
                    "quality": quality,
                }
            ),
            iterable=self.__spot_info,
            num_workers=min(num_workers, len(self.__spot_info)),
            num_retries=num_retries,
            csv_path=output_paths["spot_metadata"],
            unreadable_path=output_paths["unreadable_spots"],
            display_progress=display_progress,
        )
        return self.__spot_metadata

    def save_tiles(
        self,
        output_dir: str,
        coordinates: List[Tuple[int, int, int, int]],
        level: int = 0,
        preprocess_metrics: Callable = F.PreprocessMetrics(),
        overwrite: bool = False,
        image_format: str = "jpeg",
        quality: int = 95,
        num_workers: int = os.cpu_count(),
        num_retries: int = 10,
        display_progress: bool = True,
    ):
        """Save tile images from a slide.

        Args:
            output_dir: Output directory for the processed images.
            coordinates: A list of XYWH coordinates for the tile regions.
                Can be easily created with the `SlideReader.get_tile_coordinates()`
                method. **Please note that the level used to create coordinates
                should, match the `level` argument for this function.**
            level: Slide level to read tile images from. This should match the
                level used to create the coordinates! Defaults to 0.
            preprocess_metrics: Any Callable which takes an image as an input
                and returns a dictionary. It is recommended to use the
                `histoprep.PreprocessMetrics` class, which can be modified with
                custom functions. Defaults to `histoprep.PreprocessMetrics()`.
            overwrite: Overwrite any existing images. Defaults to False.
            image_format: Image format. Defaults to "jpeg".
            quality: Quality of JPEG compression. Defaults to 95.
            num_workers: Number of image saving workers. Defaults to the number
                of CPU cores.
            num_retries: Number retries loading any regions, which couldn't be
                saved the first time. Defaults to 10.
            display_progress: Display progress bar. Defaults to True.

        Returns:
            Tile metadata and preprocessing metrics.
        """
        # Check coordinates.
        if len(coordinates) < 1:
            warnings.warn("Empty coordinates.")
            return
        # Prepare paths and number of workers.
        num_workers, output_paths = self.__prepare_for_saving_images(
            output_dir=output_dir,
            overwrite=overwrite,
            image_format=image_format,
            quality=quality,
            num_workers=num_workers,
        )
        if output_paths is None:
            # Directory exists.
            warnings.warn(
                "Output directory exists but overwrite=False. " "Not saving images."
            )
            return
        # Create output directory.
        os.makedirs(output_paths["tile_directory"], exist_ok=True)
        self.__verbose("Saving thumbnails, tissue mask and slide summary.")
        # Save thumbnail and mask.
        self.thumbnail.save(output_paths["thumbnail"])
        self.tissue_mask.save(output_paths["tissue_mask"])
        # Save annotated thumbnail.
        annotate_thumbnail(
            thumbnail=self.thumbnail,
            thumbnail_downsample=self.thumbnail_downsample,
            coordinates=coordinates,
            coordinate_downsample=self.level_downsamples[level],
        ).save(output_paths["annotated_tiles"])
        # Save images.
        self.__tile_metadata = self.__save_images(
            worker_function=functools.partial(
                save_tile_worker,
                **{
                    "level": level,
                    "tile_directory": output_paths["tile_directory"],
                    "tissue_threshold": self.tissue_threshold,
                    "preprocess_metrics": preprocess_metrics,
                    "image_format": image_format,
                    "quality": quality,
                }
            ),
            iterable=coordinates,
            num_workers=num_workers,
            num_retries=num_retries,
            csv_path=output_paths["tile_metadata"],
            unreadable_path=output_paths["unreadable_tiles"],
            display_progress=display_progress,
        )
        return self.__tile_metadata

    def __repr__(self):
        return (
            "{}:"
            "\n  Dimensions: {}"
            "\n  Downsamples: {}"
            "\n  Channel order: {}"
            "\n  Dimension order: {}"
            "\n  Tissue threshold: {}"
            "\n  Backend: {}".format(
                self.slide_name,
                self.dimensions,
                [2**level for level in self.level_downsamples.keys()],
                self.channel_order,
                self.dimension_order,
                self.tissue_threshold,
                self.backend,
            )
        )



## DEPRACATED (for now at least)

# import logging
# import warnings
# from typing import Dict, Tuple

# import bioformats
# import numpy

# from ._base import Backend
# from ...helpers._javabridge import start_javabridge

# __all__ = ["BioformatsBackend"]

# BIOFORMATS_READABLE = bioformats.READABLE_FORMATS


# class BioformatsBackend(Backend):
#     def __init__(self, path: str):
#         # Start javabridge.
#         start_javabridge()
#         self.reader = bioformats.get_image_reader("reader", path=path)
#         # Figure out levels and dimensions.
#         self.__get_level_info()

#     def __get_level_info(self):
#         # Find out dimensions for each level.
#         level_dimensions = {}
#         for d in range(self.reader.rdr.getSeriesCount()):
#             # Read something from dimension
#             __ = self.reader.read(series=d, XYWH=(0, 0, 1, 1))
#             # Get level dimensions.
#             level_dimensions[d] = (
#                 self.reader.rdr.getSizeY(),
#                 self.reader.rdr.getSizeX(),
#             )
#         # Find out level downsamples.
#         level_downsamples = {}
#         remove = []
#         for level, dims in level_dimensions.items():
#             downsample_found = False
#             # Test different downsamples.
#             for d in range(20):
#                 # Calculate difference in dimensions form optimal downsample.
#                 y_diff = abs(dims[0] - level_dimensions[0][0] / 2**d)
#                 x_diff = abs(dims[1] - level_dimensions[0][1] / 2**d)
#                 if x_diff <= 1 and y_diff <= 1:
#                     # Differences less than one are caused by rounding.
#                     level_downsamples[level] = (
#                         level_dimensions[0][0] / dims[0],
#                         level_dimensions[0][1] / dims[1],
#                     )
#                     downsample_found = True
#                     break
#             # If we couldn't find a downsample the level does not contain
#             # an image of the tissue --> add to remove list.
#             if not downsample_found:
#                 remove.append(level)
#         # Remove levels with some other images.
#         for level in remove:
#             logging.debug(
#                 "Level {} is not a downsample of the slide image.".format(level)
#             )
#             level_dimensions.pop(level)
#         # Cahce results.
#         self.__level_dimensions = level_dimensions
#         self.__level_downsamples = level_downsamples

#     def get_dimensions(self) -> Dict[int, Tuple[int, int]]:
#         return self.__level_dimensions[0]

#     def get_level_dimensions(self) -> Dict[int, Tuple[int, int]]:
#         return self.__level_dimensions

#     def get_level_downsamples(self) -> Dict[int, Tuple[int, int]]:
#         return self.__level_downsamples

#     def get_thumbnail(self, level: int) -> numpy.ndarray:
#         # Read thumbnail.
#         thumbnail = self.reader.read(series=level)
#         # To uint8.
#         return (thumbnail * 255).astype(numpy.uint8)

#     def read_region(
#         self, XYWH: Tuple[int, int, int, int], level: int
#     ) -> numpy.ndarray:
#         # Read region
#         tile = self.reader.read(series=level, XYWH=XYWH)
#         # To uint8.
#         return (tile * 255).astype(numpy.uint8)

#     def __repr__(self):
#         return "BIOFORMATS"



import io
import sys

from histoprep.helpers._verbose import progress_bar


class CaptureStdout(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = io.StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio  # free up some memory
        sys.stdout = self._stdout


def test_progress_bar():
    with CaptureStdout() as output:
        for logs, __ in progress_bar([0, 1], desc="test", log_values=True):
            logs["dog"] = "good_boi"
    assert output == [
        "test |          | 0/2 [00:00<???]",
        "                                 ",
        "test |#####     | 1/2 [00:00<00:00]",
        "                                   ",
        "test |##########| 2/2 [00:00<00:00, dog=good_boi]",
    ]



import os
import warnings
from os.path import basename, dirname, exists, join
from typing import List, Tuple

import pandas

from ._multiprocess import multiprocess_loop
from ._verbose import progress_bar

METRIC_PREFIXES = {
    "red",
    "green",
    "blue",
    "gray",
    "saturation",
    "hue",
    "brightness",
    "sharpness",
    "black",
    "white",
}

__all__ = ["combine_metadata", "strip_metric_colums"]


def strip_metric_colums(metadata: pandas.DataFrame) -> pandas.DataFrame:
    """Remove columns with preprocessing metrics.

    Args:
        metadata: Tile metadata.

    Returns:
        Tile metadata without preprocessing metrics.

    Example:
        ```python
        from histoprep.helpers import combine_metadata, strip_metric_colums

        metadata = combine_metadata("/output/dir")
        tile_info = strip_metric_columns(metadata)
        ```
    """
    keep = []
    for col in metadata.columns:
        prefix = col.split("_")[0]
        if prefix not in METRIC_PREFIXES:
            keep.append(col)
    return metadata[keep]


def __yield_paths(parent_dir: str, filenames: List[str]):
    """Helper function for yielding metadata fileapaths."""
    for f in os.scandir(parent_dir):
        if f.is_dir():
            for name in filenames:
                path = join(f.path, name)
                if exists(path):
                    yield path


def combine_metadata(
    parent_dir: str, filename: str = "tile_metadata.csv"
) -> pandas.DataFrame:
    """Combine metadata under `parent_dir` into a single dataframe.

    Args:
        parent_dir: Output directory with the processed slides.
        filename: Filename to match. Defaults to "tile_metadata.csv".

    Raises:
        IOError: Parent directory does not exist.
        NotADirectoryError: Parent directory is not a directory.

    Returns:
        pandas.DataFrame: Combined metadata.

     Example:
        ```python
        from histoprep.helpers import combine_metadata

        metadata = combine_metadata("/output/dir")
        ```
    """
    if not exists(parent_dir):
        raise IOError("Directory {} does not exist.".format(parent_dir))
    elif os.path.isfile(parent_dir):
        raise NotADirectoryError("{} is not a directory.".format(parent_dir))
    # Collect metadata.
    combined = []
    for metadata in progress_bar(
        multiprocess_loop(__load_metadata, __yield_paths(parent_dir, [filename])),
        desc="Combining metadata",
    ):
        if metadata is not None:
            combined.append(metadata)
    if len(combined) == 0:
        return pandas.DataFrame()
    else:
        return pandas.concat(combined).reset_index(drop=True)


def __load_metadata(path: str):
    """Parallelizable metadata loading."""
    try:
        return pandas.read_csv(path)
    except Exception as e:
        warnings.warn("Could not load {} due to exception: {}.".format(path, e))
        return None


def rename_paths(parent_dir: str) -> List[Tuple[str, Exception]]:
    """Finds all metadata files inside `parent_dir` and updates the `path`
    column to match the current directory. Useful if you rename/move around
    the directories with the processed data.

    Args:
        parent_dir: Output directory with the processed slides.

    Returns:
        List csv-paths and Exceptions, for the dataframes which could not be
        updated.

     Example:
        ```python
        from histoprep.helpers import rename_paths

        failures = rename_paths("/new/output/dir")
        ```
    """
    if not exists(parent_dir):
        raise IOError("Directory {} does not exist.".format(parent_dir))
    elif os.path.isfile(parent_dir):
        raise NotADirectoryError("{} is not a directory.".format(parent_dir))
    # Rename paths in metadata
    failures = []
    for path, exception in progress_bar(
        multiprocess_loop(
            __rename_paths,
            __yield_paths(parent_dir, ["tile_metadata.csv", "spot_metadata.csv"]),
        ),
        desc="Renaming paths",
    ):
        if isinstance(exception, Exception):
            failures.append((path, exception))
    return failures


def __rename_paths(path: tuple):
    """Parallelizable path renaming."""
    try:
        # Define new directory.
        if basename(path) == "spot_metadata.csv":
            new_dir = join(dirname(path), "spots")
        elif basename(path) == "tile_metadata.csv":
            new_dir = join(dirname(path), "tiles")
        # Load metadata.
        metadata = pandas.read_csv(path)
        # Rename paths.
        metadata["path"] = [join(new_dir, basename(x)) for x in metadata["path"]]
        # Save new metadata.
        metadata.to_csv(path, index=False)
        return None, None
    except Exception as e:
        return path, e



# from ._bioformats import BioformatsBackend, BIOFORMATS_READABLE
from ._openslide import OPENSLIDE_READABLE, OpenSlideBackend
from ._pillow import PILLOW_READABLE, PillowBackend
from ._zeiss import ZEISS_READABLE, ZeissBackend

# Create readable formats
READABLE_FORMATS = tuple(set(PILLOW_READABLE + OPENSLIDE_READABLE + ZEISS_READABLE))
